from .cv import Term, TermSet


class ActivationInformation(object):
    """Describes the dissociation process used to produce an MSn scan.

    Attributes
    ----------
    data : :class:`dict`
        Additional metadata describing the dissociation
    energy : float
        The quantity of energy used
    method : :class:`DissociationMethod`
        The dissociation method used
    """

    def __init__(self, method, energy, data=None):  # pylint: disable=redefined-outer-name
        if data is None:
            data = dict()
        self.method = dissociation_methods_map.get(str(method).lower(), None)
        if self.method is None:
            self.method = self._make_unknown_method(method)
        self.energy = energy
        self.data = data

    def to_dict(self):
        d = {
            "method": str(self.method),
            "energy": self.energy,
            "data": self.data
        }
        return d

    @classmethod
    def from_dict(cls, d):
        if 'methods' in d and 'energies' in d:
            return MultipleActivationInformation(**d)
        else:
            return ActivationInformation(**d)

    @staticmethod
    def _make_unknown_method(method):  # pylint: disable=redefined-outer-name
        return DissociationMethod(UnknownDissociation.name, method, method, method, [method])

    def __repr__(self):
        return "ActivationInformation(%r, %r%s)" % (
            str(self.method), self.energy,
            "" if not self.data else ", %r" % self.data)

    def __str__(self):
        return str(self.method)

    def __hash__(self):
        return hash(str(self))

    def is_multiple_dissociation(self):
        """Determine if multiple dissociation methods were
        used.

        Returns
        -------
        bool
        """
        return False

    def has_supplemental_dissociation(self):
        """Determine if supplemental activation was used

        Returns
        -------
        bool
        """
        return False

    def __eq__(self, other):
        if other is None:
            return False
        if self.is_multiple_dissociation() != other.is_multiple_dissociation():
            return False
        if self.energy is not None:
            if other.energy is not None:
                energy_match = abs(self.energy - other.energy) < 1e-3
            else:
                energy_match = False
        else:
            if other.energy is not None:
                energy_match = False
            else:
                energy_match = True
        return (self.method == other.method) and energy_match

    def __ne__(self, other):
        return not self == other

    def has_dissociation_type(self, dissociation_type):
        return self.method.is_a(dissociation_type)


class MultipleActivationInformation(ActivationInformation):
    """Describes the dissociation processes used to produce an MSn
    scan with multiple types of activation.

    Attributes
    ----------
    data : :class:`dict`
        Additional metadata describing the dissociation
    energies : :class:`list` of float
        A list of dissociation energies used. Parallel to :attr:`methods`
    methods : :class:`list` of :class:`DissociationMethod`
        A list of dissociation methods used. Parallel to :attr:`energies`
    """

    def __init__(self, methods, energies, data=None):  # pylint: disable=super-init-not-called
        if data is None:
            data = dict()
        self.methods = []
        for method in methods:  # pylint: disable=redefined-outer-name
            self.methods.append(
                dissociation_methods_map.get(
                    str(method).lower(), method))
        self.energies = list(energies)
        self.data = data

    def to_dict(self):
        d = {
            "methods": [str(m) for m in self.methods],
            "energies": self.energies,
            "data": self.data
        }
        return d

    @property
    def method(self):
        return self.methods[0]

    @property
    def energy(self):
        return self.energies[0]

    def __iter__(self):
        return iter(self.methods)

    def __repr__(self):
        return "MultipleActivationInformation(methods=%r, energies=%r%s)" % (
            list(map(str, self.methods)), self.energies,
            "" if not self.data else ", %r" % self.data)

    def __str__(self):
        return ', '.join(map(str, self.methods))

    def is_multiple_dissociation(self):
        return True

    def has_supplemental_dissociation(self):
        return any([d.is_supplemental() for d in self.methods])

    def __eq__(self, other):
        if other is None:
            return False
        if self.is_multiple_dissociation() != other.is_multiple_dissociation():
            return False
        if len(self.energies) != len(other.energies):
            return False
        energies_equal = []
        for se, oe in zip(self.energies, other.energies):
            if se is None and oe is None:
                energies_equal.append(True)
            elif se is None or oe is None:
                energies_equal.append(False)
            else:
                energies_equal.append(abs(se - oe) < 1e-3)
        return (self.methods == other.methods) and all(energies_equal)

    def has_dissociation_type(self, dissociation_type):
        return any(method.is_a(dissociation_type) for method in self.methods)


class DissociationMethod(Term):
    """Controlled vocabulary describing a dissociation process
    """

    def is_supplemental(self):
        """Returns whether or not the dissociation process is
        supplemental

        Returns
        -------
        bool
        """
        return "supplemental" in self.name


dissociation_methods = []


# [[[cog
# import cog
# from ms_deisotope.data_source.metadata.cv import render_list
# render_list('dissociation method', term_cls_name="DissociationMethod", writer=cog.out)
# ]]]
dissociation_methods = TermSet([
    DissociationMethod(u'sustained off-resonance irradiation', u'MS:1000282',
                       (u'A technique associated with Fourier transform ion cyclotron'
                        u'resonance (FT-ICR) mass spectrometry to carry out'
                        u'ion/neutral reactions such as low-energy collision-induced'
                        u'dissociation. A radio-frequency electric field of slightly'
                        u'off-resonance to the cyclotron frequency of the reactant ion'
                        u'cyclically accelerates and decelerates the reactant ion that'
                        u"is confined in the Penning ion trap. The ion's orbit does"
                        u'not exceed the dimensions of ion trap while the ion'
                        u'undergoes an ion/neutral species process that produces a'
                        u'high average translational energy for an extended time.'),
                       'dissociation method',
                       [u'dissociation method']),
    DissociationMethod(u'post-source decay', u'MS:1000135',
                       (u'A technique specific to reflectron time-of-flight mass'
                        u'spectrometers where product ions of metastable transitions'
                        u'or collision-induced dissociations generated in the drift'
                        u'tube prior to entering the reflectron are m/z separated to'
                        u'yield product ion spectra.'),
                       'dissociation method',
                       [u'dissociation method']),
    DissociationMethod(u'plasma desorption', u'MS:1000134',
                       (u'The ionization of material in a solid sample by bombarding'
                        u'it with ionic or neutral atoms formed as a result of the'
                        u'fission of a suitable nuclide, typically 252Cf. Synonymous'
                        u'with fission fragment ionization.'),
                       'dissociation method',
                       [u'dissociation method']),
    DissociationMethod(u'surface-induced dissociation', u'MS:1000136',
                       (u'Fragmentation that results from the collision of an ion with'
                        u'a surface.'),
                       'dissociation method',
                       [u'dissociation method']),
    DissociationMethod(u'collision-induced dissociation', u'MS:1000133',
                       (u'The dissociation of an ion after collisional excitation. The'
                        u'term collisional-activated dissociation is not recommended.'),
                       'dissociation method',
                       [u'dissociation method']),
    DissociationMethod(u'pulsed q dissociation', u'MS:1000599',
                       (u'A process that involves precursor ion activation at high Q,'
                        u'a time delay to allow the precursor to fragment, then a'
                        u'rapid pulse to low Q where all fragment ions are trapped.'
                        u'The product ions can then be scanned out of the ion trap and'
                        u'detected.'),
                       'dissociation method',
                       [u'dissociation method']),
    DissociationMethod(u'electron transfer dissociation', u'MS:1000598',
                       (u'A process to fragment ions in a mass spectrometer by'
                        u'inducing fragmentation of cations (e.g. peptides or'
                        u'proteins) by transferring electrons to them.'),
                       'dissociation method',
                       [u'dissociation method']),
    DissociationMethod(u'in-source collision-induced dissociation', u'MS:1001880',
                       (u'The dissociation of an ion as a result of collisional'
                        u'excitation during ion transfer from an atmospheric pressure'
                        u'ion source and the mass spectrometer vacuum.'),
                       'dissociation method',
                       [u'dissociation method']),
    DissociationMethod(u'infrared multiphoton dissociation', u'MS:1000262',
                       (u'Multiphoton ionization where the reactant ion dissociates as'
                        u'a result of the absorption of multiple infrared photons.'),
                       'dissociation method',
                       [u'dissociation method']),
    DissociationMethod(u'blackbody infrared radiative dissociation', u'MS:1000242',
                       (u'A special case of infrared multiphoton dissociation wherein'
                        u'excitation of the reactant ion is caused by absorption of'
                        u'infrared photons radiating from heated blackbody'
                        u'surroundings, which are usually the walls of a vacuum'
                        u'chamber. See also infrared multiphoton dissociation.'),
                       'dissociation method',
                       [u'dissociation method']),
    DissociationMethod(u'low-energy collision-induced dissociation', u'MS:1000433',
                       (u'A collision-induced dissociation process wherein the'
                        u'precursor ion has the translational energy lower than'
                        u'approximately 1000 eV. This process typically requires'
                        u'multiple collisions and the collisional excitation is'
                        u'cumulative.'),
                       'dissociation method',
                       [u'dissociation method']),
    DissociationMethod(u'photodissociation', u'MS:1000435',
                       (u'A process wherein the reactant ion is dissociated as a'
                        u'result of absorption of one or more photons.'),
                       'dissociation method',
                       [u'dissociation method']),
    DissociationMethod(u'LIFT', u'MS:1002000',
                       (u"A Bruker's proprietary technique where molecular ions are"
                        u'initially accelerated at lower energy, then collide with'
                        u"inert gas in a collision cell that is then 'lifted' to high"
                        u'potential. The use of inert gas is optional, as it could'
                        u'lift also fragments provided by LID."'
                        u'[DOI:10.1007/s00216-003-2057-0'),
                       'dissociation method',
                       [u'dissociation method']),
    DissociationMethod(u'Electron-Transfer/Higher-Energy Collision Dissociation (EThcD)', u'MS:1002631',
                       (u'A dissociation process combining electron-transfer and'
                        u'higher-energy collision dissociation (EThcD). It combines'
                        u'ETD (reaction time) followed by HCD (activation energy).'),
                       'dissociation method',
                       [u'dissociation method']),
    DissociationMethod(u'electron capture dissociation', u'MS:1000250',
                       (u'A process in which a multiply protonated molecules interacts'
                        u'with a low energy electrons. Capture of the electron leads'
                        u'the liberation of energy and a reduction in charge state of'
                        u'the ion with the production of the (M + nH) (n-1)+ odd'
                        u'electron ion, which readily fragments.'),
                       'dissociation method',
                       [u'dissociation method']),
    DissociationMethod(u'trap-type collision-induced dissociation', u'MS:1002472',
                       (u'A collision-induced dissociation process that occurs in a'
                        u'trap-type collision cell.'),
                       'dissociation method',
                       [u'collision-induced dissociation', u'dissociation method']),
    DissociationMethod(u'beam-type collision-induced dissociation', u'MS:1000422',
                       (u'A collision-induced dissociation process that occurs in a'
                        u'beam-type collision cell.'),
                       'dissociation method',
                       [u'collision-induced dissociation', u'dissociation method']),
    DissociationMethod(u'supplemental collision-induced dissociation', u'MS:1002679',
                       (u'The dissociation of an ion after supplemental collisional'
                        u'excitation.'),
                       'dissociation method',
                       [u'collision-induced dissociation', u'dissociation method']),
    DissociationMethod(u'higher energy beam-type collision-induced dissociation', u'MS:1002481',
                       (u'A collision-induced dissociation process wherein the'
                        u'projectile ion has the translational energy higher than'
                        u'approximately 1000 eV.'),
                       'dissociation method',
                       [u'beam-type collision-induced dissociation', u'collision-induced dissociation', u'dissociation method']),
    DissociationMethod(u'supplemental beam-type collision-induced dissociation', u'MS:1002678',
                       (u'A supplemental collision-induced dissociation process that'
                        u'occurs in a beam-type collision cell in addition to another'
                        u'primary type of dissociation.'),
                       'dissociation method',
                       [u'beam-type collision-induced dissociation', u'collision-induced dissociation', u'dissociation method']),
])
# [[[end]]]


UnknownDissociation = DissociationMethod(
    "unknown dissociation", None, None, 'dissociation method',
    [u'dissociation method'])


dissociation_methods_map = {
    "": UnknownDissociation,
    None: UnknownDissociation,
    UnknownDissociation.name: UnknownDissociation
}


method = None
for method in dissociation_methods:
    dissociation_methods_map[method.name] = method
del method


CID = dissociation_methods_map.get("collision-induced dissociation")
HCD = dissociation_methods_map.get("beam-type collision-induced dissociation")
ETD = dissociation_methods_map.get("electron transfer dissociation")
ECD = dissociation_methods_map.get("electron capture dissociation")

supplemental_term_map = {
    dissociation_methods_map["beam-type collision-induced dissociation"]: dissociation_methods_map[
        "supplemental beam-type collision-induced dissociation"],
    dissociation_methods_map["collision-induced dissociation"]: dissociation_methods_map[
        "supplemental collision-induced dissociation"]
}


supplemental_energy = "supplemental collision energy"


energy_terms = set([
    "collision energy",
    supplemental_energy,
    "activation energy",
    "collision energy ramp start",
    "collision energy ramp end",
    "percent collision energy ramp start",
    "percent collision energy ramp end",
])


dissociation_methods_map.update({
    'cad': CID,
    'cid': CID,
    "hcd": HCD,
    "etd": ETD,
    "ecd": ECD
})


ActivationInformation.dissociation_methods = dissociation_methods_map


__all__ = [
    "ActivationInformation", "MultipleActivationInformation", "DissociationMethod",
    "dissociation_methods", "UnknownDissociation", "dissociation_methods_map",
    "CID", "HCD", "ETD", "ECD", "supplemental_energy", "supplemental_term_map",
    "energy_terms"
]



'''Defines proxy objects which appear to be :class:`~.Scan` objects without
issueing any I/O operations until a scan's attributes are requested, and the a
LRU cache for keeping only recently used full scans in memory. This allows
a large number of scans to be "loaded" at once without requiring storing all of
the backing information.
'''
import weakref
import logging
import operator

from collections import OrderedDict

from .scan import ScanBase

UNLOAD_POLICY_FULL = "unload_policy_full"
UNLOAD_POLICY_KEEP = "unload_policy_keep"

logger = logging.getLogger("ms_deisotope.scan_proxy")
logger.addHandler(logging.NullHandler())

LOAD_METHOD_ID = 'id'
LOAD_METHOD_INDEX = 'index'


class ScanProxyContext(object):
    """A memory-conserving wrapper around an existing :class:`RandomAccessScanSource`
    object for serving :class:`ScanProxy` objects.

    Attributes
    ----------
    cache : :class:`collections.OrderedDict`
        A strong-reference maintaining cache from scan ID to :class:`ScanBase`
    cache_size : int
        The number of scans to keep strong references to.
    source : :class:`RandomAccessScanSource`
        The source to load scans from.
    """

    def __init__(self, source, cache_size=2 ** 10, unload_policy=None, track_allocations=False):
        if unload_policy is None:
            unload_policy = UNLOAD_POLICY_FULL
        self.source = source
        self.cache_size = cache_size
        self.cache = OrderedDict()
        self.unload_policy = unload_policy
        self.track_allocations = track_allocations
        self.allocation_map = weakref.WeakValueDictionary()

    def allocation_statistics(self):
        churn = {}
        for proxy in self.allocation_map.values():
            churn[proxy._target_scan_id] = proxy._unload_count
        return churn

    def _refresh(self, key):
        value = self.cache.pop(key)
        self.cache[key] = value
        return value

    def get_scan_by_id(self, scan_id):
        '''Retrieve a real scan by its identifier.

        This method first checks the in-memory cache for the scan identifier
        and returns it if already loaded. If not, the scan is retrieved from
        :attr:`source`, usually from disk and saved to the cache using
        :meth:`_save_scan`. This may cause the cache eviction.

        Parameters
        ----------
        scan_id: :class:`str`
            The scan to retrieve

        Returns
        -------
        :class:`~.Scan`
        '''
        if scan_id in self.cache:
            return self._refresh(scan_id)
        scan = self.source.get_scan_by_id(scan_id)
        self._save_scan(scan_id, scan)
        return scan

    def get_scan_by_index(self, index):
        '''Retrieve a real scan by its index.

        This method first checks the in-memory cache for the scan index
        and returns it if already loaded. If not, the scan is retrieved from
        :attr:`source`, usually from disk and saved to the cache using
        :meth:`_save_scan`. This may cause the cache eviction.

        Parameters
        ----------
        index: :class:`int`
            The scan to retrieve

        Returns
        -------
        :class:`~.Scan`
        '''
        if index in self.cache:
            return self._refresh(index)
        scan = self.source.get_scan_by_index(index)
        self._save_scan(index, scan)
        return scan

    def _save_scan(self, scan_id, scan):
        if len(self.cache) > self.cache_size:
            _, evicted_scan = self.cache.popitem(last=False)
            self.source._scan_cleared(evicted_scan)
        self.cache[scan_id] = scan

    def __call__(self, scan_id, method=LOAD_METHOD_ID):
        """Forward call to :meth:`create_scan_proxy`.

        Parameters
        ----------
        scan_id: :class:`str`

        Returns
        -------
        :class:`ScanProxy`

        See Also
        --------
        :meth:`create_scan_proxy`
        """
        return self.create_scan_proxy(scan_id, method)

    def create_scan_proxy(self, scan_id, method=LOAD_METHOD_ID):
        """Create a proxy for the scan referenced by scan_id

        Parameters
        ----------
        scan_id: :class:`str`

        Returns
        -------
        :class:`ScanProxy`
        """
        proxy = ScanProxy(scan_id, self, method=method)
        if self.track_allocations:
            self.allocation_map[scan_id] = proxy
        return proxy

    def clear(self):
        '''Clear the reference cache.

        If the scans that were previously held in the cache were not
        strong-referenced anywhere else, any extant proxies which depend
        upon this context will have to reload their spectra from disk.
        '''
        self.cache.clear()


class proxyproperty(object):
    '''An descriptor that integrates with :class:`ScanProxy` to retrieve attributes
    from a wrapped :class:`~.ScanBase` object referenced by :class:`ScanProxy`,
    potentially loading the scan from disk if it has been purged from the memory
    cache.

    Attributes
    ----------
    name: :class:`str`
        The name of the attribute to retrieve from the wrapped scan
    '''

    def __init__(self, name, caching=False):
        self.name = name
        self.caching = caching
        self.is_null_slot = "_%s_null" % self.name
        self.cache_slot = "_%s" % self.name
        self._null_getter = operator.attrgetter(self.is_null_slot)
        self._cache_getter = operator.attrgetter(self.cache_slot)
        self._name_getter = operator.attrgetter(self.name)

    def _refresh(self, instance):
        setattr(instance, self.is_null_slot, None)

    def _clear(self, instance):
        setattr(instance, self.is_null_slot, None)
        setattr(instance, self.cache_slot, None)

    def __get__(self, instance, owner):
        if self.caching:
            is_null_slot = self.is_null_slot
            cache_slot = self.cache_slot
            try:
                # is_null = getattr(instance, is_null_slot)
                is_null = self._null_getter(instance)
            except AttributeError:
                setattr(instance, is_null_slot, None)
                is_null = None
            if is_null:
                return None
            try:
                # value = getattr(instance, cache_slot)
                value = self._cache_getter(instance)
                has_value = value is not None
            except AttributeError:
                has_value = False
            if not has_value:
                # value = getattr(instance.scan, self.name)
                value = self._name_getter(instance.scan)
                if is_null is None:
                    if value is None:
                        setattr(instance, is_null_slot, True)
                    else:
                        setattr(instance, is_null_slot, False)
                setattr(instance, cache_slot, value)
            return value
        # return getattr(instance.scan, self.name)
        return self._name_getter(instance.scan)

    def __set__(self, instance, value):
        if self.caching:
            setattr(instance, self.cache_slot, value)
            setattr(instance, self.is_null_slot, None)
        else:
            raise TypeError("Cannot set attribute \"%s\"" % (self.name, ))

    def __delete__(self, instance):
        if self.caching:
            self._clear(instance)


class ScanProxy(ScanBase):
    """A proxy for a :class:`~.ScanBase` object, allowing transparent access to the
    scan, potentially loading it from disk through the attached :class:`ScanProxyContext`
    :attrib:`context`.

    Attributes
    ----------
    context : :class:`ScanProxyContext`
        The context to use to retrieve scans from, whose cache layer will keep the
        scan objects alive.
    scan : :class:`weakref.ProxyType`
        A weakref proxy to the :class:`~.ScanBase` object.
    """

    _names = [
        "arrays",
        "id",
        "title",
        "ms_level",
        "scan_time",
        "index",
        "is_profile",
        "polarity",
        "activation",
        "acquisition_information",
        "isolation_window",
        "instrument_configuration",
        "annotations",
        "tic",
        "base_peak",
    ]

    def __init__(self, scan_id, context, method=LOAD_METHOD_ID):
        self._target_scan_id = scan_id
        self.context = context
        self.load_method = method
        self._scan = None
        self._precursor_information = None
        self._peak_set = None
        self._unload_count = 0

    @property
    def source(self):
        '''The source of the scan data this proxy is bound to.

        Returns
        -------
        :class:`RandomAccessScanSource`
        '''
        return self.context.source

    def _clear_scan(self, *args, **kwargs):
        self._scan = None
        self._unload_count += 1
        logger.debug("Unloading Scan %s | %d", self._target_scan_id, self._unload_count)
        if self.context.unload_policy == UNLOAD_POLICY_FULL:
            self._peak_set = None
            self._deconvoluted_peak_set = None

    def _load_scan(self):
        scan = None
        if self.load_method == LOAD_METHOD_ID:
            scan = self.context.get_scan_by_id(self._target_scan_id)
        elif self.load_method == LOAD_METHOD_INDEX:
            scan = self.context.get_scan_by_index(self._target_scan_id)
        else:
            raise TypeError(
                "Cannot resolve scan loading method {!r}".format(self.load_method))
        return scan

    def _require_scan(self):
        if self._scan is None:
            scan = self._load_scan()
            self._scan = weakref.proxy(
                scan,
                self._clear_scan)
        else:
            # remind :attr:`context`'s cache that this scan was used.
            self._load_scan()


    @property
    def scan(self):
        '''The proxied scan.

        Accessing this property may require loading the scan's data
        from disk and/or triggering a cache eviction.
        '''
        self._require_scan()
        return self._scan

    def pick_peaks(self, *args, **kwargs):
        '''Request the proxied scan picked peaks if they are not
        already available and then cache them in memory.
        '''
        self._require_scan()
        peaks = self.peak_set
        if peaks is None:
            peaks = self.scan.pick_peaks(*args, **kwargs).peak_set
        self.peak_set = peaks
        return self

    def _resolve_sequence(self):
        deconvoluted_peak_set = self.deconvoluted_peak_set
        if deconvoluted_peak_set is not None:
            return deconvoluted_peak_set
        peak_set = self.peak_set
        if peak_set is not None:
            return peak_set
        self.pick_peaks()
        return self.peak_set

    def __getitem__(self, i):
        return self._resolve_sequence()[i]

    def __iter__(self):
        return iter(self._resolve_sequence())

    def __len__(self):
        return len(self._resolve_sequence())

    def has_peak(self, *args, **kwargs):
        '''Query the wrapped scan's peaks using it's :meth:`Scan.has_peak`
        method. If no peaks are available, this will call :meth:`pick_peaks`
        first to resolve the peak set and then query its peaks instead.

        Parameters
        ----------
        mz: float
            The m/z to search for
        error_tolerance: float
            The parts per million mass error tolerance to use

        Returns
        -------
        ms_peak_picker.FittedPeak or None
            The peak closest to the query m/z within the error tolerance window or None
            if there are no peaks satisfying the requirements

        See Also
        --------
        :meth:`Scan.has_peak`
        '''
        try:
            return self.scan.has_peak(*args, **kwargs)
        except ValueError:
            return self.pick_peaks().has_peak(*args, **kwargs)

    peak_set = proxyproperty('peak_set', True)
    deconvoluted_peak_set = proxyproperty('deconvoluted_peak_set', True)
    precursor_information = proxyproperty("precursor_information", True)

    def __repr__(self):
        template = "{self.__class__.__name__}({self._target_scan_id!r})"
        return template.format(self=self)

    @classmethod
    def _configure_proxy_attributes(cls):
        for name in cls._names:
            setattr(cls, name, proxyproperty(name))

    def _refresh(self):
        self.__class__.peak_set._refresh(self)
        self.__class__.deconvoluted_peak_set._refresh(self)
        self.__class__.precursor_information._refresh(self)


ScanProxy._configure_proxy_attributes()



from libc.math cimport sqrt, exp, pi

cimport cython
import numpy as np
cimport numpy as np

np.import_array()


cdef:
    double sqrt2pi = sqrt(2 * pi)
    double sqrt2 = sqrt(2)
    double SIGMA_EPSILON = 1e-3


cdef double erf(double x):
    '''https://www.johndcook.com/blog/2009/01/19/stand-alone-error-function-erf/'''
    # constants
    cdef double a1 =  0.254829592
    cdef double a2 = -0.284496736
    cdef double a3 =  1.421413741
    cdef double a4 = -1.453152027
    cdef double a5 =  1.061405429
    cdef double p  =  0.3275911

    # Save the sign of x
    cdef int sign = 1
    if x < 0:
        sign = -1
    x = abs(x)

    # A & S 7.1.26
    cdef double t = 1.0 / (1.0 + p * x)
    cdef double y = 1.0 - ((((( a5 * t + a4) * t) + a3) * t + a2) * t + a1) * t * exp(-x * x)

    return sign*y


@cython.boundscheck(False)
@cython.cdivision(True)
cpdef skewed_gaussian_shape(np.ndarray[np.float64_t, ndim=1] xs, double center,
                            double amplitude, double sigma, double gamma):
    cdef:
        np.ndarray[np.float64_t, ndim=1] temp
        size_t i, n
        double sigma_squared_2, xi, t, amp_over_sigma_sqrt2pi, sigma_sqrt2

    if sigma == 0:
        sigma = SIGMA_EPSILON

    sigma_squared_2 = 2 * (sigma ** (2))
    amp_over_sigma_sqrt2pi = (amplitude) / (sigma * sqrt(2 * pi))
    sigma_sqrt2 = (sigma * sqrt(2))

    temp = xs * 0

    n = temp.shape[0]
    for i in range(n):
        xi = xs[i]
        t = -((xi - center) ** 2) / (sigma_squared_2)
        temp[i] =  (amp_over_sigma_sqrt2pi * exp(t)) * (
            1 + erf((gamma * (xi - center)) / sigma_sqrt2))

    return temp

@cython.boundscheck(False)
@cython.cdivision(True)
cpdef bigaussian_shape(np.ndarray[np.float64_t, ndim=1] xs, double center,
                       double amplitude, double sigma_left, double sigma_right):
    cdef:
        np.ndarray[np.float64_t, ndim=1] temp
        size_t i, n, center_index
        double xi, t, two_sigma_left_squared, two_sigma_right_squared

    if sigma_left == 0:
        sigma_left = SIGMA_EPSILON
    if sigma_right == 0:
        sigma_right = SIGMA_EPSILON

    temp = xs * 0
    n = temp.shape[0]

    two_sigma_left_squared = (2 * sigma_left ** 2)
    two_sigma_right_squared = (2 * sigma_right ** 2)

    for i in range(n):
        xi = xs[i]
        # left side
        if xi < center:
            t = amplitude / sqrt2pi * exp(-(xi - center) ** 2 / two_sigma_left_squared)
        # right side
        else:
            t = amplitude / sqrt2pi * exp(-(xi - center) ** 2 / two_sigma_right_squared)
        temp[i] = t
    return temp


@cython.boundscheck(False)
@cython.cdivision(True)
cpdef gaussian_shape(np.ndarray[np.float64_t, ndim=1] xs, double center,
                     double amplitude, double sigma):
    cdef:
        np.ndarray[np.float64_t, ndim=1] temp
        size_t i, n
        double xi, t, two_sigma_squared

    if sigma == 0:
        sigma = SIGMA_EPSILON

    two_sigma_squared = (2 * sigma ** 2)

    temp = xs * 0
    n = temp.shape[0]

    for i in range(n):
        xi = xs[i]
        t = (amplitude / (sigma * sqrt2pi)) * exp(-(xi - center) ** 2 / two_sigma_squared)
        temp[i] = t

    return temp



import ms_deisotope

from ms_deisotope.feature_map import quick_index
from ms_deisotope.test.common import datafile


mzml_path = datafile("small.mzML")


def test_quick_index():
    reader = ms_deisotope.MSFileLoader(mzml_path)
    index, _interval_tree = quick_index.index(reader)
    n_1 = len(index.ms1_ids)
    n_n = len(index.msn_ids)
    assert n_1 == 14
    assert n_n == 34



'''Represent the basic structures of a mass spectrum and its processed contents,
and provide an interface for manipulating that data.
'''
import warnings
try:
    from collections import Sequence
except ImportError:
    from collections.abc import Sequence

import numpy as np

from ms_peak_picker import (
    pick_peaks, reprofile, average_signal,
    scan_filter, PeakIndex, PeakSet)

from ms_deisotope.utils import decimal_shift
from ms_deisotope.deconvolution import deconvolute_peaks

from ms_deisotope.data_source.metadata.scan_traits import (
    IsolationWindow,
    ScanAcquisitionInformation)

from ms_deisotope.data_source.metadata.activation import ActivationInformation
from ms_deisotope.data_source.metadata.instrument_components import InstrumentInformation


from .base import (ScanBase, RawDataArrays, PrecursorInformation)


class Scan(ScanBase):
    """Container for mass spectral data and associated descriptive information.

    A :class:`Scan` object is a generic object intended to be created by a :class:`ScanDataSource` and describes
    a mass spectrum at each level of processing (Profile --> Peak Fitted --> Deconvoluted). The raw object
    provided by the source is wrapped and queried lazily when an attribute is requested, delegated through
    :attr:`source`.

    Attributes
    ----------
    deconvoluted_peak_set : :class:`ms_deisotope.DeconvolutedPeakSet` or None
        Deconvoluted peaks resulting from charge state deconvolution and deisotoping. Will
        be `None` if deconvolution has not been done.
    peak_set : :class:`ms_peak_picker.PeakSet` or None
        Picked peaks and (possibly) associated raw data points as produced by :meth:`pick_peaks`.
        Will be `None` if peak picking has not been done.
    product_scans : list
        A list of :class:`Scan` instances which were produced by fragmenting ions from this one.
        This attribute is not guaranteed to be populated depending upon how the scan is loaded.
    source : :class:`ScanDataSource`
        The object which produced this scan and which defines the methods for retrieving common
        attributes from the underlying data structures.
    precursor_information: :class:`PrecursorInformation` or None
        Descriptive metadata for the ion which was chosen for fragmentation, and a reference to
        the precursor scan
    arrays: :class:`RawDataArrays`
        A pair of :class:`numpy.ndarray` objects corresponding to the raw m/z and intensity data points
    id: str
        The unique identifier for this scan as given by the source
    title: str
        The human-readable display string for this scan as shown in some external software
    ms_level: int
        The degree of fragmentation performed. 1 corresponds to a MS1 or "Survey" scan, 2 corresponds
        to MS/MS, and so on. If :attr:`ms_level` > 1, the scan is considered a "tandem scan" or "MS^n" scan
    scan_time: float
        The time the scan was acquired during data acquisition. The unit of time will always be minutes.
    drift_time: float or None
        The time measured by the ion mobility spectrometer for this scan or frame. This quantity is None
        if the scan does not have ion mobility information associated with it, which is usually recorded
        in :attr:`acquisition_information`
    index: int
        The integer number indicating how many scans were acquired prior to this scan.
    is_profile: bool
        Whether this scan's raw data points corresponds to a profile scan or whether the raw data was
        pre-centroided.
    polarity: int
        If the scan was acquired in positive mode, the value ``+1``.  If the scan was acquired in negative
        mode, the value ``-1``. May be used to indicating how to calibrate charge state determination methods.
    activation: :class:`.ActivationInformation` or None
        If this scan is an MS^n scan, this attribute will contain information about the process
        used to produce it from its parent ion.
    instrument_configuration: :class:`~.InstrumentInformation`
        The instrument configuration used to acquire this scan.
    acquisition_information: :class:`.ScanAcquisitionInformation` or None
        Describes the type of event that produced this scan, as well as the scanning method
        used.
    isolation_window: :class:`.IsolationWindow` or None
        Describes the range of m/z that were isolated from a parent scan to create this scan
    annotations: dict
        A set of key-value pairs describing the scan not part of the standard interface
    """

    def __init__(self, data, source, peak_set=None, deconvoluted_peak_set=None, product_scans=None, annotations=None):
        if product_scans is None:
            product_scans = []
        if annotations is None:
            annotations = dict()
        self.source = source
        self.peak_set = peak_set
        self.deconvoluted_peak_set = deconvoluted_peak_set

        self._data = data

        self._arrays = None
        self._id = None
        self._title = None
        self._ms_level = None
        self._scan_time = None
        self._precursor_information = None
        self._index = None
        self._is_profile = None
        self._polarity = None
        self._activation = None
        self._acquisition_information = None
        self._isolation_window = None
        self._instrument_configuration = None

        self._annotations = None
        self._external_annotations = annotations

        self.product_scans = product_scans

    def clone(self, deep=True):
        """Return a copy of the :class:`Scan` object
        wrapping the same reference data, potentially a deep
        one

        Parameters
        ----------
        deep: :class:`bool`

        Returns
        -------
        :class:`Scan`
        """
        dup = self.__class__(
            self._data, self.source,

            self.peak_set.clone() if self.peak_set is not None else None
            if deep else self.peak_set,

            self.deconvoluted_peak_set.clone() if self.deconvoluted_peak_set is not None else None
            if deep else self.deconvoluted_peak_set,

            [s.clone(deep=deep) for s in self.product_scans], self._external_annotations.copy())
        return dup

    def _load(self):
        _ = self.arrays
        _ = self.id
        _ = self.title
        _ = self.ms_level
        _ = self.scan_time
        _ = self.index
        _ = self.polarity
        _ = self.precursor_information
        _ = self.activation
        _ = self.acquisition_information
        _ = self.isolation_window
        _ = self.is_profile
        _ = self.instrument_configuration
        _ = self.annotations
        _ = None
        del _
        return self

    def _unload(self):
        self._arrays = None
        self._id = None
        self._title = None
        self._ms_level = None
        self._scan_time = None
        self._precursor_information = None
        self._index = None
        self._is_profile = None
        self._polarity = None
        self._activation = None
        self._acquisition_information = None
        self._isolation_window = None
        self._instrument_configuration = None

    def clear(self):
        """Releases all associated in-memory data and clears the cached
        attributes.

        The data reference attribute :attr:`_data` is retained
        and unchanged.
        """
        if self.source is not None:
            self.source._scan_cleared(self)
        self._unload()

    @property
    def ms_level(self):
        '''The degree of fragmentation performed. 1 corresponds to a MS1 or "Survey" scan, 2 corresponds
        to MS/MS, and so on. If :attr:`ms_level` > 1, the scan is considered a "tandem scan" or "MS^n" scan

        Returns
        -------
        :class:`int`
        '''
        if self._ms_level is None:
            self._ms_level = self.source._ms_level(self._data)
        return self._ms_level

    @ms_level.setter
    def ms_level(self, value):
        self._ms_level = int(value)

    @property
    def is_profile(self):
        '''Whether this scan's raw data points corresponds to a profile scan or whether the raw data was
        pre-centroided.

        Returns
        -------
        :class:`bool`
        '''
        if self._is_profile is None:
            self._is_profile = self.source._is_profile(self._data)
        return self._is_profile

    @is_profile.setter
    def is_profile(self, value):
        self._is_profile = bool(value)

    @property
    def polarity(self):
        '''If the scan was acquired in positive mode, the value ``+1``.  If the scan was acquired in negative
        mode, the value ``-1``. May be used to indicating how to calibrate charge state determination methods.

        Returns
        -------
        :class:`int`
        '''
        if self._polarity is None:
            self._polarity = self.source._polarity(self._data)
        return self._polarity

    @polarity.setter
    def polarity(self, value):
        self._polarity = int(value)

    @property
    def scan_time(self):
        '''The time the scan was acquired during data acquisition. The unit of time will always
        be minutes.

        Returns
        -------
        :class:`float`
        '''
        if self._scan_time is None:
            self._scan_time = self.source._scan_time(self._data)
        return self._scan_time

    @scan_time.setter
    def scan_time(self, value):
        self._scan_time = float(value)

    @property
    def arrays(self):
        '''A pair of :class:`numpy.ndarray` objects corresponding to the raw m/z and
        intensity data points.

        These arrays are wrapped in a :class:`~.RawDataArrays` instance, which provides
        additional methods.

        Returns
        -------
        :class:`~.RawDataArrays`
        '''
        if self._arrays is None:
            self._arrays = RawDataArrays(*self.source._scan_arrays(self._data))
        return self._arrays

    @arrays.setter
    def arrays(self, value):
        if isinstance(value, RawDataArrays) or value is None:
            self._arrays = value
        elif isinstance(value, Sequence):
            if len(value) == 2:
                self._arrays = RawDataArrays(*map(np.asanyarray, value))
            elif len(value) == 3:
                RawDataArrays(*map(np.asanyarray, value[:2]), arrays=dict(value[2]))
        else:
            raise TypeError(
                "arrays must be an instance of RawDataArrays or a pair of numpy arrays")

    @property
    def title(self):
        '''The human-readable display string for this scan as shown in some external software.

        Returns
        -------
        :class:`str`
        '''
        if self._title is None:
            self._title = self.source._scan_title(self._data)
        return self._title

    @title.setter
    def title(self, value):
        self._title = value

    @property
    def id(self):
        '''The within run unique scan identifier.

        Returns
        -------
        :class:`str`
        '''
        if self._id is None:
            self._id = self.source._scan_id(self._data)
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    scan_id = id

    @property
    def index(self):
        '''The integer number indicating how many scans were acquired prior to this scan.

        Returns
        -------
        :class:`int`
        '''
        if self._index is None:
            self._index = self.source._scan_index(self._data)
        return self._index

    @index.setter
    def index(self, value):
        self._index = int(value)

    @property
    def precursor_information(self):
        '''Descriptive metadata for the ion which was chosen for fragmentation, and a reference to
        the precursor scan.

        Returns
        -------
        :class:`~.PrecursorInformation`'''
        if self.ms_level < 2:
            return None
        if self._precursor_information is None:
            self._precursor_information = self.source._precursor_information(
                self._data)
        return self._precursor_information

    @precursor_information.setter
    def precursor_information(self, value):
        if not isinstance(value, PrecursorInformation) and value is not None:
            raise TypeError("precursor_information must be a %r instance" % (
                PrecursorInformation, ))
        self._precursor_information = value

    @property
    def activation(self):
        '''If this scan is an MS^n scan, this attribute will contain information about the process
        used to produce it from its parent ion.

        Returns
        -------
        :class:`~.ActivationInformation`
        '''
        if self.ms_level < 2:
            return None
        if self._activation is None:
            self._activation = self.source._activation(self._data)
        return self._activation

    @activation.setter
    def activation(self, value):
        if not isinstance(value, ActivationInformation) and value is not None:
            raise TypeError(
                "activation must be an %r instance" % (ActivationInformation, ))
        self._activation = value

    @property
    def isolation_window(self):
        '''Describes the range of m/z that were isolated from a parent scan to create this scan.

        Returns
        -------
        :class:`~.IsolationWindow`
        '''
        if self.ms_level < 2:
            return None
        if self._isolation_window is None:
            self._isolation_window = self.source._isolation_window(self._data)
        return self._isolation_window

    @isolation_window.setter
    def isolation_window(self, value):
        if isinstance(value, IsolationWindow) or value is None:
            self._isolation_window = value
        elif isinstance(value, Sequence):
            if len(value) == 2:
                lo, hi = value
                width = (hi - lo) / 2.
                center = lo + width
                self._isolation_window = IsolationWindow(center, width, width)
            elif len(value) == 3:
                lo, center, hi = value
                self._isolation_window = IsolationWindow(lo, center, hi)
            else:
                raise ValueError("Could not convert %r to an %r" %
                                 (value, IsolationWindow))
        else:
            raise TypeError(
                "isolation_window must be an either an %r instance or a sequence of two or three elements" % (
                    IsolationWindow))

    @property
    def acquisition_information(self):
        '''Describes the type of event that produced this scan, as well as the scanning method
        used.'''
        if self._acquisition_information is None:
            self._acquisition_information = self.source._acquisition_information(
                self._data)
        return self._acquisition_information

    @acquisition_information.setter
    def acquisition_information(self, value):
        if not isinstance(value, ScanAcquisitionInformation) and value is not None:
            raise TypeError("acquisition_information must be an instance of %r" % (
                ScanAcquisitionInformation, ))
        self._acquisition_information = value

    @property
    def instrument_configuration(self):
        '''The instrument configuration used to acquire this scan.'''
        if self._instrument_configuration is None:
            self._instrument_configuration = self.source._instrument_configuration(
                self._data)
        return self._instrument_configuration

    @instrument_configuration.setter
    def instrument_configuration(self, value):
        if not isinstance(value, InstrumentInformation) and value is not None:
            raise TypeError("instrument_configuration must be an instance of %r" % (
                InstrumentInformation, ))
        self._instrument_configuration = value

    @property
    def annotations(self):
        '''A set of key-value pairs describing the scan not part of the standard interface'''
        if self._annotations is None:
            self._annotations = self.source._annotations(self._data)
            self._annotations.update(self._external_annotations)
        return self._annotations

    @annotations.setter
    def annotations(self, value):
        self._external_annotations = dict(value)
        self._annotations = self._external_annotations.copy()

    def bind(self, source):
        super(Scan, self).bind(source)
        self.source = source
        return self

    def unbind(self):
        super(Scan, self).unbind()
        self.source = None
        return self

    def __repr__(self):
        try:
            precursor_information = self.precursor_information
        except KeyError:
            precursor_information = None
        return "Scan(%r, index=%d, time=%0.4f, ms_level=%r%s)" % (
            self.id, (self.index if self.index is not None else -1), (
                self.scan_time if self.scan_time is not None else -1), self.ms_level,
            ", " + repr(precursor_information) if precursor_information else '')

    # peak manipulation

    def __iter__(self):
        if self.peak_set is None:
            raise ValueError("Cannot iterate over peaks in a scan that has not been "
                             "centroided. Call `pick_peaks` first.")
        return iter(self.peak_set)

    def __getitem__(self, i):
        if self.peak_set is None:
            raise ValueError("Cannot retrieve peaks in a scan that has not been "
                             "centroided. Call `pick_peaks` first.")
        return self.peak_set[i]

    def __len__(self):
        if self.peak_set is None:
            raise ValueError("Cannot count peaks in a scan that has not been "
                             "centroided. Call `pick_peaks` first.")
        return len(self.peak_set)

    def __bool__(self):
        return bool(self._data)

    def __nonzero__(self):
        return self.__bool__()

    def has_peak(self, *args, **kwargs):
        """A wrapper around :meth:`ms_peak_picker.PeakSet.has_peak` to query the
        :class:`ms_peak_picker.FittedPeak` objects picked for this scan.

        Parameters
        ----------
        mz: float
            The m/z to search for
        error_tolerance: float
            The parts per million mass error tolerance to use

        Returns
        -------
        :class:`ms_peak_picker.FittedPeak` or None
            The peak closest to the query m/z within the error tolerance window or None
            if there are no peaks satisfying the requirements

        Raises
        ------
        ValueError:
            If the scan has not yet had peaks picked yet

        See Also
        --------
        :meth:`.Scan.pick_peaks`
        """
        if self.peak_set is None:
            raise ValueError("Cannot search for peaks in a scan that has not been "
                             "centroided. Call `pick_peaks` first.")
        return self.peak_set.has_peak(*args, **kwargs)

    def pick_peaks(self, *args, **kwargs):
        """A wrapper around :func:`ms_peak_picker.pick_peaks` which will populate the
        :attr:`peak_set` attribute of this scan.

        Parameters
        ----------
        fit_type : str, optional
            The name of the peak model to use. One of "quadratic", "gaussian", "lorentzian", or "apex"
        signal_to_noise_threshold : int, optional
            Minimum signal-to-noise measurement to accept a peak
        intensity_threshold : int, optional
            Minimum intensity measurement to accept a peak
        threshold_data : bool, optional
            Whether to apply thresholds to the data
        target_envelopes : list, optional
            A sequence of (start m/z, end m/z) pairs, limiting peak picking to only those intervals
        transforms : list, optional
            A list of :class:`scan_filter.FilterBase` instances or callable that
            accepts (mz_array, intensity_array) and returns (mz_array, intensity_array) or
            `str` matching one of the premade names in `scan_filter.filter_register`
        verbose : bool, optional
            Whether to log extra information while picking peaks
        start_mz : float, optional
            A minimum m/z value to start picking peaks from
        stop_mz : float, optional
            A maximum m/z value to stop picking peaks after
        *args :
            Passed along to :func:`ms_peak_picker.pick_peaks`
        **kwargs :
            Passed along to :func:`ms_peak_picker.pick_peaks`

        Returns
        -------
        Scan
            Returns self
        """
        # Check to see if the user requested one of the ms_peak_picker fits or wanted
        # to use the vendor peak picker if provided.
        fit_type_k = kwargs.get("fit_type")
        if len(args) > 0:
            fit_type_a = args[0]
        else:
            fit_type_a = None
        if fit_type_k == 'vendor' or fit_type_a == 'vendor':
            try:
                peaks = self.source._pick_peaks_vendor(self._data, *args, **kwargs)
                self.peak_set = peaks
                return self
            except NotImplementedError:
                pass
        # Prepare the peak picking parameters
        mzs, intensities = self.arrays
        if len(mzs) == 0:
            self.peak_set = PeakIndex(mzs, intensities, PeakSet([]))
            return self
        if self.is_profile:
            peak_mode = 'profile'
        else:
            peak_mode = 'centroid'

        kwargs.setdefault('peak_mode', peak_mode)

        self.peak_set = pick_peaks(mzs, intensities, *args, **kwargs)
        return self

    def deconvolute(self, *args, **kwargs):
        """A wrapper around :func:`ms_deisotope.deconvolution.deconvolute_peaks`.

        The scan must have had its peaks picked before it can be deconvoluted.

        Parameters
        ----------
        decon_config : dict, optional
            Parameters to use to initialize the deconvoluter instance produced by
            ``deconvoluter_type``
        charge_range : tuple of integers, optional
            The range of charge states to consider.
        error_tolerance : float, optional
            PPM error tolerance to use to match experimental to theoretical peaks
        priority_list : list, optional
            The set of peaks to target for deconvolution to be able to enforce external
            constraints on, such as selected precursors for fragmentation.
        left_search_limit : int, optional
            The maximum number of neutron shifts to search to the left  (decrease) from
            each query peak
        right_search_limit : int, optional
            The maximum number of neutron shifts to search to the right (increase) from
            each query peak
        left_search_limit_for_priorities : int, optional
            The maximum number of neutron shifts to search to the left (decrease) from
            each query peak for priority targets
        right_search_limit_for_priorities : None, optional
            The maximum number of neutron shifts to search to the right (increase) from
            each query peak for priority targets
        charge_carrier : float, optional
            The mass of the charge carrier. Defaults to PROTON
        truncate_after : float, optional
            The percentage of the isotopic pattern to include. Defaults to TRUNCATE_AFTER
        deconvoluter_type : type or callable, optional
            A callable returning a deconvoluter. Defaults to :class:`~.AveraginePeakDependenceGraphDeconvoluter`
        **kwargs
            Additional keywords passed to :func:`~.deconvolute_peaks`

        Returns
        -------
        Scan
            Returns self

        Raises
        ------
        ValueError
            If :attr:`peak_set` is None, a :class:`ValueError` will be raised
            indicating that a scan must be centroided before it can be deconvoluted

        See Also
        --------
        :func:`~.deconvolute_peaks`
        """
        if self.peak_set is None:
            raise ValueError("Cannot deconvolute a scan that has not been "
                             "centroided. Call `pick_peaks` first.")
        charge_range = kwargs.get("charge_range", (1, 8))
        if self.polarity < 0 and max(charge_range) > 0:
            charge_range = tuple(c * self.polarity for c in charge_range)
        kwargs['charge_range'] = charge_range
        decon_results = deconvolute_peaks(self.peak_set, *args, **kwargs)
        self.deconvoluted_peak_set = decon_results.peak_set
        return self

    def pack(self):
        '''Pack the (dispersed) representation of the data in this :class:`Scan`
        into a packed :class:`ProcessedScan` object.

        Returns
        -------
        :class:`ProcessedScan`
        '''
        precursor_info = self.precursor_information
        return ProcessedScan(
            self.id, self.title, precursor_info,
            self.ms_level, self.scan_time, self.index,
            self.peak_set.pack() if self.peak_set is not None else None,
            self.deconvoluted_peak_set,
            self.polarity,
            self.activation,
            self.acquisition_information,
            self.isolation_window,
            self.instrument_configuration,
            self.product_scans,
            self.annotations)

    # signal transformation

    def reprofile(self, max_fwhm=0.2, dx=0.01, model_cls=None):
        """Use the picked peaks in :attr:`peak_set` to create a new
        profile mass spectrum using a peak shape model.

        Parameters
        ----------
        max_fwhm : float, optional
            Maximum peak width above which peaks will be ignored
        dx : float, optional
            The distance between each new point in m/z space in the
            reprofiled spectrum
        model_cls : ms_peak_picker.peak_statistics.PeakShapeModel, optional
            The peak shape model to use to generate the profile data from
            the centroided peaks. Defaults a Gaussian model

        Returns
        -------
        Scan
            A shallow copy of this scan with its :attr:`arrays` replaced with
            the new reprofiled arrays

        Raises
        ------
        ValueError
            A scan that has not been centroided and is already in profile mode
            must have its peaks picked before it can be reprofiled.
        """
        if self.peak_set is None and self.is_profile:
            raise ValueError(
                "Cannot reprofile a scan that has not been centroided")
        elif self.peak_set is None and not self.is_profile:
            self.pick_peaks()
        if not self.peak_set:
            arrays = (np.array([], dtype=float), np.array([], dtype=float))
        else:
            arrays = reprofile(self.peak_set, max_fwhm, dx, model_cls)
        scan = WrappedScan(
            self._data, self.source, arrays,
            list(self.product_scans), is_profile=True,
            annotations=self._external_annotations)
        return scan

    def denoise(self, scale=5.0, window_length=2.0, region_width=10):
        """Create a shallow copy of the scan with a noise reduction
        transformation applied.

        This method uses the scan filter :class:`ms_peak_picker.scan_filter.FTICRBaselineRemoval`
        which uses the MasSpike noise reduction algorithm.

        Parameters
        ----------
        scale : float, optional
            The multiplier of the local noise window to remove
        window_length : float, optional
            The width (in m/z) of each window
        region_width : int, optional
            The width (in m/z) of each region of windows

        Returns
        -------
        Scan
            The denoised version of this scan
        """
        mzs, intensities = self.arrays
        mzs = mzs.astype(float)
        intensities = intensities.astype(float)
        transform = scan_filter.FTICRBaselineRemoval(
            window_length=window_length, scale=scale, region_width=region_width)
        mzs, intensities = transform(mzs, intensities)
        return WrappedScan(self._data, self.source,
                           (mzs, intensities), list(self.product_scans),
                           is_profile=self.is_profile,
                           annotations=self._external_annotations)

    def transform(self, filters=None):
        '''Applies a series of :class:`ms_peak_picker.scan_filter.FilterBase`,
        or strings that are recognized by :func:`ms_peak_picker.scan_filter.transform`

        Arguments
        ---------
        filters: :class:`Iterable`
            An iterable of transformations of :class:`ms_peak_picker.scan_filter.FilterBase`
            or strings.

        Returns
        -------
        :class:`WrappedScan`
        '''
        mzs, intensities = self.arrays
        mzs = mzs.astype(float)
        intensities = intensities.astype(float)
        mzs, intensities = scan_filter.transform(
            mzs, intensities, filters=filters)
        return WrappedScan(self._data, self.source,
                           (mzs, intensities), list(self.product_scans),
                           is_profile=self.is_profile,
                           annotations=self._external_annotations)

    def average_with(self, scans, dx=None, weight_sigma=None):
        r"""Average together multiple scans' raw data arrays to create a composite intensity
        profile for a common m/z axis.

        Parameters
        ----------
        scans: list:
            A list of :class:`Scan` objects
        dx : float, optional
            The distance between each point in the generated common m/z axis.
        weight_sigma : float, optional
            When this value is not None, scans are weighted according to a
            gaussian distribution with a $\sigma$ equal to this value

        Returns
        -------
        :class:`AveragedScan`
            A shallow copy of this scan with its :attr:`arrays` attribute replaced
            with the averaged array
        """
        if dx is None:
            dx = 0.01
            default_dx = True
        else:
            default_dx = False
        scans = [self] + list(scans)
        arrays = []
        for scan in scans:
            if scan.is_profile:
                arrays.append(scan.arrays)
            else:
                arrays.append(scan.reprofile(dx=dx).arrays)
        if weight_sigma:
            if weight_sigma == 1:
                weight_sigma = 0.025
            weights = self._compute_smoothing_weights(
                scans, mean=self.scan_time, sigma=weight_sigma)
        else:
            weights = None
        if default_dx:
            if len(arrays) > 2:
                reference = arrays[len(arrays) // 2 + 1]
            else:
                reference = arrays[0]
            empirical_dx = decimal_shift(2 * np.median(np.diff(reference.mz)))
            dx = min(dx, empirical_dx)
        new_arrays = average_signal(arrays, dx=dx, weights=weights)
        indices = [scan.index for scan in scans]
        return AveragedScan(
            self._data, self.source, new_arrays,
            indices, list(self.product_scans), is_profile=True,
            annotations=self._external_annotations)

    def _get_adjacent_scans(self, index_interval=None, rt_interval=None):
        if index_interval is None and rt_interval is None:
            raise ValueError(
                "One of `index_interval` or `rt_interval` must be provided")
        if self.ms_level > 1:
            raise ValueError("Cannot average MSn scans at this time")
        if not self.source:
            raise ValueError("Can't average an unbound scan")
        before = []
        after = []
        if index_interval is not None:
            before = []
            current_index = self.index
            for _ in range(index_interval):
                next_scan = self.source.find_previous_ms1(current_index)
                if next_scan is None:
                    break
                before.append(next_scan)
                current_index = next_scan.index
            before = before[::-1]
            after = []
            current_index = self.index
            for _ in range(index_interval):
                try:
                    next_scan = self.source.find_next_ms1(current_index)
                except ValueError:
                    break
                if next_scan is None:
                    break
                after.append(next_scan)
                current_index = next_scan.index
        elif rt_interval is not None:
            reference_time = self.scan_time
            before = []
            current_index = self.index
            current_time = self.scan_time
            while abs(reference_time - current_time) < rt_interval and current_index > 0:
                next_scan = self.source.find_previous_ms1(current_index)
                if next_scan is None:
                    break
                before.append(next_scan)
                current_index = next_scan.index
                current_time = next_scan.scan_time

            before = before[::-1]

            after = []
            current_index = self.index
            current_time = self.scan_time
            while abs(reference_time - current_time) < rt_interval and current_index > 0:
                try:
                    next_scan = self.source.find_next_ms1(current_index)
                except ValueError:
                    break
                if next_scan is None:
                    break
                after.append(next_scan)
                current_index = next_scan.index
                current_time = next_scan.scan_time
        else:
            raise ValueError(
                "One of `index_interval` or `rt_interval` must be provided")
        return before, after

    def _compute_smoothing_weights(self, scans, mean, sigma=0.025):
        sigma_sqrd_2 = (2 * sigma ** 2)
        time_array = np.array([s.scan_time for s in scans])
        weights = np.exp((-(time_array - mean) ** 2) / sigma_sqrd_2)
        return weights

    def average(self, index_interval=None, rt_interval=None, dx=None, weight_sigma=None):
        r"""Average together multiple scans' raw data arrays to create a composite intensity
        profile for a common m/z axis.

        Only MS1 scans will be averaged with this method

        Either an absolute number of scans before and after can be specified using
        ``index_interval`` or a time window may be specified using ``rt_interval``.

        Parameters
        ----------
        index_interval : int, optional
            The number of scans preceding and proceding to average with.
        rt_interval : float, optional
            The range of time (in minutes) preceding and proceding to
            look for other scans to average with.
        dx : float, optional
            The distance between each point in the generated common m/z axis.
        weight_sigma : float, optional
            When this value is not None, scans are weighted according to a
            gaussian distribution with a $\sigma$ equal to this value

        Returns
        -------
        :class:`AveragedScan`
            A shallow copy of this scan with its :attr:`arrays` attribute replaced
            with the averaged array
        """
        if dx is None:
            dx = 0.01
            default_dx = True
        else:
            default_dx = False
        before, after = self._get_adjacent_scans(index_interval, rt_interval)
        scans = before + [self] + after
        arrays = []
        for scan in scans:
            if scan.is_profile:
                scan_arrays = scan.arrays
            else:
                scan_arrays = scan.reprofile(dx=dx).arrays
            if len(scan_arrays.mz) > 0:
                arrays.append(scan_arrays)
        if weight_sigma:
            if weight_sigma == 1:
                weight_sigma = 0.025
            weights = self._compute_smoothing_weights(
                scans, mean=self.scan_time, sigma=weight_sigma)
        else:
            weights = None
        if default_dx:
            if len(arrays) > 2:
                reference = arrays[len(arrays) // 2 + 1]
            else:
                reference = arrays[0]
            empirical_dx = decimal_shift(2 * np.median(np.diff(reference.mz)))
            dx = min(dx, empirical_dx)

        new_arrays = average_signal(arrays, dx=dx, weights=weights)
        indices = [scan.index for scan in scans]
        return AveragedScan(
            self._data, self.source, new_arrays,
            indices, list(self.product_scans), is_profile=True,
            annotations=self._external_annotations)


class WrappedScan(Scan):
    '''A wrapper around a :class:`Scan` object with one or more attributes overridden.
    '''
    overridable_keys = [
        "_arrays",
        "_id",
        "_title",
        "_ms_level",
        "_scan_time",
        "_precursor_information",
        "_index",
        "_is_profile",
        "_polarity",
        "_activation",
        "_acquisition_information",
        "_isolation_window",
        "_instrument_configuration"
    ]

    def __init__(self, data, source, array_data, product_scans=None, annotations=None, **overrides):
        super(WrappedScan, self).__init__(
            data, source, peak_set=None,
            deconvoluted_peak_set=None,
            annotations=annotations,
            product_scans=product_scans)
        self._arrays = RawDataArrays(*array_data)
        self._overrides = overrides
        for key, value in overrides.items():
            if not key.startswith("_"):
                key = "_" + key
            if key in self.overridable_keys:
                setattr(self, key, value)
            else:
                warnings.warn("Cannot override attribute %s" % (key,))

    def clone(self, deep=True):
        dup = self.__class__(
            self._data, self.source, self.arrays,
            [s.clone(deep=deep) for s in self.product_scans],
            annotations=self._external_annotations,
            **self._overrides)
        if deep:
            dup.peak_set = self.peak_set.clone() if self.peak_set is not None else None
            dup.deconvoluted_peak_set = self.deconvoluted_peak_set.clone()\
                if self.deconvoluted_peak_set is not None else None
        else:
            dup.peak_set = self.peak_set
            dup.deconvoluted_peak_set = self.deconvoluted_peak_set
        return dup


class AveragedScan(WrappedScan):
    '''An averaged :class:`Scan` object, storing additional information for retrieving
    the spectra that were averaged together.

    Attributes
    ----------
    scan_indices: list
        The :attr:`index` values for all of the scans that were averaged together
    '''

    def __init__(self, data, source, array_data, scan_indices, product_scans=None, annotations=None, **overrides):
        super(AveragedScan, self).__init__(
            data, source, array_data,
            product_scans=product_scans, annotations=annotations, **overrides)
        self.scan_indices = scan_indices

    def clone(self, deep=True):
        dup = self.__class__(
            self._data, self.source, self.arrays,
            self.scan_indices,
            [s.clone(deep=deep) for s in self.product_scans],
            annotations=self._external_annotations,
            **self._overrides)
        if deep:
            dup.peak_set = self.peak_set.clone() if self.peak_set is not None else None
            dup.deconvoluted_peak_set = self.deconvoluted_peak_set.clone()\
                if self.deconvoluted_peak_set is not None else None
        else:
            dup.peak_set = self.peak_set
            dup.deconvoluted_peak_set = self.deconvoluted_peak_set
        return dup


class ProcessedScan(ScanBase):
    """Container for mass spectral data and associated descriptive information that has been
    processed and reduced.

    A :class:`ProcessedScan` object has been processed and compacted. It does not carry a
    :class:`~.RawDataArrays` :attr:`arrays` attribute, only at least one of :attr:`peak_set`
    and :attr:`deconvoluted_peak_set`. The rest of its attributes are all loaded in memory
    directly without needing to reconstruct them from a specific format lazily through a
    :class:`~.ScanDataSource`.

    Attributes
    ----------
    deconvoluted_peak_set : :class:`~.DeconvolutedPeakSet` or None
        Deconvoluted peaks resulting from charge state deconvolution and deisotoping. Will
        be `None` if deconvolution has not been done.
    peak_set : :class:`ms_peak_picker.PeakSet` or None
        Picked peaks and (possibly) associated raw data points as produced by :meth:`pick_peaks`.
        Will be `None` if peak picking has not been done.
    product_scans : list
        A list of :class:`Scan` instances which were produced by fragmenting ions from this one.
        This attribute is not guaranteed to be populated depending upon how the scan is loaded.
    source : :class:`ScanDataSource`
        The object which produced this scan and which defines the methods for retrieving common
        attributes from the underlying data structures.
    precursor_information: :class:`PrecursorInformation` or None
        Descriptive metadata for the ion which was chosen for fragmentation, and a reference to
        the precursor scan
    id: str
        The unique identifier for this scan as given by the source
    title: str
        The human-readable display string for this scan as shown in some external software
    ms_level: int
        The degree of fragmentation performed. 1 corresponds to a MS1 or "Survey" scan, 2 corresponds
        to MS/MS, and so on. If :attr:`ms_level` > 1, the scan is considered a "tandem scan" or "MS^n" scan
    scan_time: float
        The time the scan was acquired during data acquisition. The unit of time will always be minutes.
    drift_time: float or None
        The time measured by the ion mobility spectrometer for this scan or frame. This quantity is None
        if the scan does not have ion mobility information associated with it, which is usually recorded
        in :attr:`acquisition_information`
    index: int
        The integer number indicating how many scans were acquired prior to this scan.
    is_profile: bool
        Whether this scan's raw data points corresponds to a profile scan or whether the raw data was
        pre-centroided.
    polarity: int
        If the scan was acquired in positive mode, the value ``+1``.  If the scan was acquired in negative
        mode, the value ``-1``. May be used to indicating how to calibrate charge state determination methods.
    activation: :class:`.ActivationInformation` or None
        If this scan is an MS^n scan, this attribute will contain information about the process
        used to produce it from its parent ion.
    instrument_configuration: :class:`~.InstrumentInformation`
        The instrument configuration used to acquire this scan.
    acquisition_information: :class:`.ScanAcquisitionInformation` or None
        Describes the type of event that produced this scan, as well as the scanning method
        used.
    isolation_window: :class:`.IsolationWindow` or None
        Describes the range of m/z that were isolated from a parent scan to create this scan
    annotations: dict
        A set of key-value pairs describing the scan not part of the standard interface
    """

    def __init__(self, id, title, precursor_information,
                 ms_level, scan_time, index, peak_set,
                 deconvoluted_peak_set, polarity=None, activation=None,
                 acquisition_information=None, isolation_window=None,
                 instrument_configuration=None, product_scans=None,
                 annotations=None):
        if product_scans is None:
            product_scans = []
        if annotations is None:
            annotations = {}
        self.id = id
        self.title = title
        self.precursor_information = precursor_information
        self.ms_level = ms_level
        self.scan_time = scan_time
        self.index = index
        self.peak_set = peak_set
        self.deconvoluted_peak_set = deconvoluted_peak_set
        self.polarity = polarity
        self.activation = activation
        self.acquisition_information = acquisition_information
        self.isolation_window = isolation_window
        self.instrument_configuration = instrument_configuration
        self.product_scans = product_scans
        self.annotations = annotations
        self.source = None

    def clear(self):
        '''Clear storage-heavy attribute values
        '''
        self.peak_set = None
        self.deconvoluted_peak_set = None
        self.activation = None
        self.acquisition_information = None
        self.isolation_window = None
        self.instrument_configuration = None
        self.product_scans = None

    @property
    def scan_id(self):
        return self.id

    @property
    def is_profile(self):
        '''Whether this scan's raw data points corresponds to a profile scan or whether the raw data was
        pre-centroided.
        '''
        return False

    def _resolve_peaks(self):
        if self.deconvoluted_peak_set is not None:
            return self.deconvoluted_peak_set
        elif self.peak_set is not None:
            return self.peak_set
        else:
            return []

    def __iter__(self):
        return iter(self._resolve_peaks())

    def __getitem__(self, index):
        return self._resolve_peaks()[index]

    def __len__(self):
        return len(self._resolve_peaks())

    def has_peak(self, mass, error_tolerance=2e-5):
        """A wrapper around :meth:`~.DeconvolutedPeakSet.has_peak` to query the
        :class:`~.DeconvolutedPeak` objects picked for this scan. If no deconvoluted
        peaks are available, but centroided peaks are, this method will instead
        behave like :class:`Scan.has_peak`

        Parameters
        ----------
        mass: float
            The mass to search for
        error_tolerance: float
            The parts per million mass error tolerance to use

        Returns
        -------
        :class:`~.PeakBase`
        """
        if self.deconvoluted_peak_set is not None:
            return self.deconvoluted_peak_set.has_peak(mass, error_tolerance)
        elif self.peak_set is not None:
            return self.peak_set.has_peak(mass, error_tolerance)
        else:
            raise ValueError("No peaks available")

    def __repr__(self):
        if self.deconvoluted_peak_set is not None:
            peaks = self.deconvoluted_peak_set
        elif self.peak_set is not None:
            peaks = self.peak_set
        else:
            peaks = []

        pinfo = self.precursor_information
        if pinfo:
            pinfo_string = ", %s" % pinfo
        else:
            pinfo_string = ""

        return "ProcessedScan(id=%s, ms_level=%d, %d peaks%s)" % (
            self.id, self.ms_level, len(peaks), pinfo_string)

    def bind(self, source):
        super(ProcessedScan, self).bind(source)
        self.source = source
        return self

    def unbind(self):
        super(ProcessedScan, self).unbind()
        self.source = None
        return self

    def clone(self, deep=True):
        """Return a copy of the :class:`ProcessedScan` object, potentially a deep
        one

        Parameters
        ----------
        deep: :class:`bool`

        Returns
        -------
        :class:`Scan`
        """
        dup = self.__class__(
            self.id, self.title, self.precursor_information, self.ms_level,
            self.scan_time, self.index,

            self.peak_set.clone() if self.peak_set is not None else None
            if deep else self.peak_set,

            self.deconvoluted_peak_set.clone() if self.deconvoluted_peak_set is not None else None
            if deep else self.deconvoluted_peak_set,

            self.polarity, self.activation, self.acquisition_information,
            self.isolation_window, self.instrument_configuration,
            list(self.product_scans), self.annotations.copy())
        return dup



import click

import numpy as np
from array import array

import ms_deisotope
from ms_deisotope import plot


@click.group('draw')
def draw():
    '''Draw images related to mass spectrometry data!
    '''

def _make_figure():
    try:
        from matplotlib.figure import Figure
        from matplotlib.backends.backend_agg import (FigureCanvasAgg,)
        figure = Figure(dpi=120)
        canvas = FigureCanvasAgg(figure)  # pylint: disable=unused-variable
        axis = figure.add_subplot(111)
        return figure, axis
    except ImportError:
        raise click.ClickException(
            "Could not import matplotlib, please make sure it is installed prior to using `draw` commands")


@draw.command("tic")
@click.argument("path", type=click.Path(exists=True))
@click.option("-o", "--output-path", type=click.Path(writable=True),
              required=False, help="Path to write image to")
@click.option("-s", "--start-time", type=float, required=False, help='Time to start from')
@click.option("-e", "--end-time", type=float, required=False, help='Time to end at')
def draw_tic(path, output_path=None, start_time=None, end_time=None):
    """Draw the Total Ion Chromatogram (TIC), the total signal at each time point.
    """
    if output_path is None:
        output_path = path + '.tic.png'
    if start_time is None:
        start_time = 0
    if end_time is None:
        end_time = float('inf')

    figure, axis = _make_figure()

    reader = ms_deisotope.MSFileLoader(path)
    reader.start_from_scan(rt=start_time, grouped=False)

    time = array('d')
    intensity = array('d')

    bar = click.progressbar(reader, item_show_func=lambda x: str(
        x.id) if x is not None else '')
    with bar:
        for scan in bar:
            if scan.ms_level != 1:
                continue
            time.append(scan.scan_time)
            intensity.append(scan.arrays.intensity.sum())
    click.echo("Total Ion Current: %e" % np.sum(intensity))

    axis.plot(time, intensity)
    axis.set_xlabel("Scan Time (Min)", fontsize=16)
    axis.set_ylabel("Relative Intensity", fontsize=16)
    ylim = axis.get_ylim()
    axis.set_ylim(-10, ylim[1])
    axis.set_xlim(time[0] - 2, time[-1] + 2)
    figure.text(0.15, 0.8, "%0.3e" % np.sum(intensity), ha='left')
    figure.savefig(output_path, bbox_inches='tight', dpi=120)


@draw.command("spectrum")
@click.argument("ms-file-path", type=click.Path(exists=True))
@click.option("-o", "--output-path", type=click.Path(writable=True),
              required=False, help="Path to write image to")
@click.option("-i", "--index", type=int, help="The index of the scan to draw. Exclusive with scan-id and time.")
@click.option("-s", "--scan-id", type=str, help="The unique ID of the scan to draw. Exclusive with index and time.")
@click.option("-t", "--time", type=float,
              help="The acquisition time of the scan to draw. Exclusive with index and scan-id.")
def draw_spectrum(ms_file_path, index=None, scan_id=None, time=None, output_path=None):
    options = map(bool, (index, scan_id, time))
    if sum(options) == 0 or sum(options) > 1:
        raise click.UsageError(
            "Only one of `index`, `scan-id`, and `time` should be provided")

    reader = ms_deisotope.MSFileLoader(ms_file_path)

    if index is not None:
        source = "index"
        key = index
        scan = reader.get_scan_by_index(index)
    elif scan_id is not None:
        source = 'scan_id'
        key = scan_id
        scan = reader.get_scan_by_id(scan_id)
    elif time is not None:
        source = 'time'
        key = time
        scan = reader.get_scan_by_time(time)

    click.echo("Drawing %r" % (scan, ))

    if output_path is None:
        output_path = ms_file_path + '.%s-%s.png' % (source, key)

    figure, axis = _make_figure()

    if scan.is_profile:
        scan.arrays.plot(ax=axis)
    else:
        scan.pick_peaks()
        plot.draw_peaklist(scan.peak_set, ax=axis)
    figure.savefig(output_path, bbox_inches='tight', dpi=120)



import os
import threading

import logging

try:
    from Queue import Queue, Empty as QueueEmptyException
except ImportError:
    from queue import Queue, Empty as QueueEmptyException

from ms_deisotope.data_source import MSFileLoader, ScanBunch
from ms_deisotope.data_source.metadata.file_information import (
    SourceFile as MetadataSourceFile, FileInformation)

from ms_deisotope.output.mzml import MzMLScanSerializer
from ms_deisotope.output.mgf import MGFSerializer

from ms_deisotope.task import TaskBase

DONE = b'---NO-MORE---'

logger = logging.getLogger("ms_deisotope.deisotoper.output")


class ScanStorageHandlerBase(TaskBase):
    def __init__(self, *args, **kwargs):
        self.current_precursor = None
        self.current_products = []

    def reset(self):
        self.current_precursor = None
        self.current_products = []

    def save_bunch(self, precursor, products):
        raise NotImplementedError()

    def save(self):
        if self.current_precursor is not None:
            self.save_bunch(
                self.current_precursor, self.current_products)
            self.reset()

    def register_parameter(self, name, value):
        pass

    def complete(self):
        pass

    @classmethod
    def configure_storage(cls, path=None, name=None, source=None):
        return cls()

    def accumulate(self, scan):
        if self.current_precursor is None:
            self.current_precursor = scan
        elif scan.ms_level == self.current_precursor.ms_level:
            self.save()
            self.current_precursor = scan
        else:
            self.current_products.append(scan)

    def commit(self):
        self.save()

    def sync(self):
        self.commit()

    def _get_sample_run(self):
        return None

    @property
    def sample_run(self):
        return self._get_sample_run()


class NullScanStorageHandler(ScanStorageHandlerBase):
    def save_bunch(self, precursor, products):
        pass


class ThreadedScanStorageHandlerMixin(object):
    def __init__(self, *args, **kwargs):
        queue_size = int(kwargs.get("queue_size", 200))
        self.queue = Queue(queue_size)
        self.worker_thread = threading.Thread(target=self._worker_loop)
        self.worker_thread.start()
        super(ThreadedScanStorageHandlerMixin, self).__init__(*args, **kwargs)

    def _save_bunch(self, precursor, products):
        self.serializer.save(ScanBunch(precursor, products))
        try:
            precursor.clear()
            for product in products:
                product.clear()
        except AttributeError:
            pass

    def save_bunch(self, precursor, products):
        self.queue.put((precursor, products))

    def _worker_loop(self):
        has_work = True
        i = 0

        def drain_queue():
            current_work = []
            try:
                while len(current_work) < 300:
                    current_work.append(self.queue.get_nowait())
            except QueueEmptyException:
                pass
            if len(current_work) > 50:
                self.log("Drained Write Queue of %d items" % (len(current_work),))
            return current_work

        while has_work:
            try:
                next_bunch = self.queue.get(True, 1)
                i += 1
                if next_bunch == DONE:
                    has_work = False
                    continue
                self._save_bunch(*next_bunch)
                if self.queue.qsize() > 0:
                    current_work = drain_queue()
                    for next_bunch in current_work:
                        i += 1
                        if next_bunch == DONE:
                            has_work = False
                        else:
                            self._save_bunch(*next_bunch)
                            i += 1
            except QueueEmptyException:
                continue
            except Exception as e:
                self.error("An error occurred while writing scans to disk", e)

    def sync(self):
        self._end_thread()
        self.worker_thread = threading.Thread(target=self._worker_loop)
        self.worker_thread.start()

    def _end_thread(self):
        self.queue.put(DONE)
        if self.worker_thread is not None:
            self.worker_thread.join()

    def commit(self):
        super(ThreadedScanStorageHandlerMixin, self).save()
        self._end_thread()

    def complete(self):
        self.save()
        self._end_thread()
        super(ThreadedScanStorageHandlerMixin, self).complete()


class MzMLScanStorageHandler(ScanStorageHandlerBase):
    def __init__(self, path, sample_name, n_spectra=None, deconvoluted=True):
        if n_spectra is None:
            n_spectra = 2e5
        super(MzMLScanStorageHandler, self).__init__()
        self.path = path
        self.handle = open(path, 'wb')
        self.serializer = MzMLScanSerializer(
            self.handle, n_spectra, sample_name=sample_name,
            deconvoluted=deconvoluted)

    def _get_sample_run(self):
        return self.serializer.sample_run

    def register_parameter(self, name, value):
        self.serializer.add_processing_parameter(name, value)

    @classmethod
    def configure_storage(cls, path=None, name=None, source=None):
        if path is not None:
            if name is None:
                sample_name = os.path.basename(path)
            else:
                sample_name = name
        else:
            path = "processed.mzML"
            sample_name = name if name is not None else ""
        if source is not None:
            reader = MSFileLoader(source.scan_source)
            n_spectra = len(reader.index)
            deconvoluting = source.deconvoluting
            inst = cls(path, sample_name, n_spectra=n_spectra, deconvoluted=deconvoluting)
            try:
                description = reader.file_description()
            except AttributeError:
                description = FileInformation()
            source_file_metadata = MetadataSourceFile.from_path(source.scan_source)
            inst.serializer.add_file_information(description)
            try:
                inst.serializer.remove_file_contents("profile spectrum")
            except KeyError:
                pass
            inst.serializer.add_file_contents("centroid spectrum")
            if source_file_metadata not in description.source_files:
                inst.serializer.add_source_file(source_file_metadata)
            try:
                instrument_configs = reader.instrument_configuration()
                for config in instrument_configs:
                    inst.serializer.add_instrument_configuration(config)
            except Exception as e:
                inst.error(
                    "An error occurred while writing instrument configuration", e)
            for trans in source.ms1_peak_picking_args.get("transforms", []):
                inst.register_parameter("parameter: ms1-%s" % trans.__class__.__name__, repr(trans))
            if deconvoluting:
                if source.ms1_deconvolution_args.get("averagine"):
                    inst.register_parameter(
                        "parameter: ms1-averagine", repr(source.ms1_deconvolution_args.get("averagine")))
                if source.ms1_deconvolution_args.get("scorer"):
                    inst.register_parameter(
                        "parameter: ms1-scorer", repr(source.ms1_deconvolution_args.get("scorer")))
                if source.ms1_averaging > 0:
                    inst.register_parameter("parameter: ms1-averaging", repr(source.ms1_averaging))
                if source.ignore_tandem_scans:
                    inst.register_parameter("parameter: ignore-tandem-scans", "")
                if source.extract_only_tandem_envelopes:
                    inst.register_parameter("parameter: extract-only-tandem-envelopes", "")

            if source.msn_peak_picking_args is not None:
                for trans in source.msn_peak_picking_args.get("transforms", []):
                    inst.register_parameter("parameter: msn-%s" % trans.__class__.__name__, repr(trans))
            if deconvoluting:
                if source.msn_deconvolution_args.get("averagine"):
                    inst.register_parameter(
                        "parameter: msn-averagine", repr(source.msn_deconvolution_args.get("averagine")))
                if source.msn_deconvolution_args.get("scorer"):
                    inst.register_parameter(
                        "parameter: msn-scorer", repr(source.msn_deconvolution_args.get("scorer")))
            data_processing = inst.serializer.build_processing_method()
            inst.serializer.add_data_processing(data_processing)
        else:
            n_spectra = 2e5
            inst = cls(path, sample_name, n_spectra=n_spectra)
        # Force marshalling of controlled vocabularies early.
        inst.serializer.writer.param("32-bit float")
        return inst

    def save_bunch(self, precursor, products):
        self.serializer.save_scan_bunch(ScanBunch(precursor, products))

    def complete(self):
        self.save()
        self.serializer.close()


class ThreadedMzMLScanStorageHandler(ThreadedScanStorageHandlerMixin, MzMLScanStorageHandler):
    def __init__(self, path, sample_name, n_spectra=None, deconvoluted=True):
        super(ThreadedMzMLScanStorageHandler, self).__init__(
            path, sample_name, n_spectra, deconvoluted=deconvoluted)


class MGFScanStorageHandler(ScanStorageHandlerBase):
    def __init__(self, path, sample_name, deconvoluted=True, **kwargs):
        super(MGFScanStorageHandler, self).__init__()
        self.path = path
        self.handle = open(path, "wb")
        self.serializer = MGFSerializer(
            self.handle, sample_name=sample_name,
            deconvoluted=deconvoluted)

    def _get_sample_run(self):
        return self.serializer.sample_run

    def register_parameter(self, name, value):
        self.serializer.add_global_parameter(name, value)

    @classmethod
    def configure_storage(cls, path=None, name=None, source=None):
        if path is not None:
            if name is None:
                sample_name = os.path.basename(path)
            else:
                sample_name = name
        else:
            path = "processed.mgf"
            sample_name = name if name is not None else ""
        if source is not None:
            deconvoluting = source.deconvoluting
            inst = cls(path, sample_name, deconvoluted=deconvoluting)
        else:
            inst = cls(path, sample_name)
        return inst

    def save_bunch(self, precursor, products):
        self.serializer.save_scan_bunch(ScanBunch(precursor, products))

    def complete(self):
        self.save()
        self.serializer.close()


class ThreadedMGFScanStorageHandler(ThreadedScanStorageHandlerMixin, MGFScanStorageHandler):
    def __init__(self, path, sample_name, n_spectra=None, deconvoluted=True):
        super(ThreadedMGFScanStorageHandler, self).__init__(
            path, sample_name, n_spectra, deconvoluted=deconvoluted)



# pragma: no cover

import numpy as np
from ms_peak_picker import FittedPeak, PeakIndex, PeakSet
from ms_deisotope.peak_set import DeconvolutedPeak, DeconvolutedPeakSet

from collections import Counter, defaultdict


def unspool_peaks(peak_sets):
    return [p for peak_set in peak_sets for p in peak_set]


class PeakScanTree(object):
    def __init__(self, store=None):
        if store is None:
            store = dict()
        self.store = defaultdict(list, store)
        self._mz = None

    def add(self, peak, scan_id):
        self.store[scan_id].append(peak)
        if self._mz is None:
            self._mz = peak.mz

    def extend(self, tree):
        for key, peaks in tree.store.items():
            for peak in peaks:
                self.add(peak, key)

    def __iter__(self):
        for peaks in self.store.values():
            for peak in peaks:
                yield peak

    @property
    def mz(self):
        return self._mz

    def remove_scan(self, scan_id):
        self.store.pop(scan_id)
        self._reset()

    def _reset(self):
        vals = self.store.values()
        if vals:
            peak = vals[0][0]
            self._mz = peak.mz
        else:
            self._mz = None


class PeakCluster(object):
    def __init__(self, peaks=None):
        if peaks is None:
            peaks = []
        self.peaks = peaks
        self.mz = None
        self.probability = 1

    def add(self, peak):
        if isinstance(peak, PeakCluster):
            self.peaks.extend(peak)
        else:
            self.peaks.append(peak)
        if self.mz is None:
            self.mz = self.peaks[0].mz

    def __repr__(self):
        if len(self) == 0:
            return "PeakCluster()"
        return "PeakCluster(%f, %d)" % (self.mz, len(self.peaks))

    def __iter__(self):
        return iter(self.peaks)

    def __len__(self):
        return len(self.peaks)

    def weighted_mz(self):
        prod = 0
        total_intensity = 0
        for peak in self:
            prod += peak.mz * peak.intensity
            total_intensity += peak.intensity
        return (prod / total_intensity)

    def _scaling_factor(self):
        return 0.95 + 0.05 * (1 + self.probability) ** 5

    @property
    def intensity(self):
        return sum(peak.intensity for peak in self) * self._scaling_factor()


class PeakClusterBuilder(object):
    def __init__(self, error_tolerance=1e-5):
        self.error_tolerance = error_tolerance
        self.peak_clusters = []
        self.count = 0

    def __len__(self):
        return len(self.peak_clusters)

    def __iter__(self):
        return iter(self.peak_clusters)

    def __getitem__(self, i):
        if isinstance(i, (int, slice)):
            return self.peak_clusters[i]
        else:
            return [self.peak_clusters[j] for j in i]

    def has_peak(self, mz, error_tolerance=1e-5):
        ix, matched = binary_search_with_flag(self.peak_clusters, mz, error_tolerance)
        if matched:
            return self[ix]

    def find_insertion_point(self, peak):
        index, matched = binary_search_with_flag(
            self.peak_clusters, peak.mz, self.error_tolerance)
        return index, matched

    def find_minimizing_index(self, peak, indices):
        best_index = None
        best_error = float('inf')
        for index_case in indices:
            chroma = self[index_case]
            err = abs(chroma.mz - peak.mz) / peak.mz
            if err < best_error:
                best_index = index_case
                best_error = err
        return best_index

    def handle_peak(self, peak):
        if len(self) == 0:
            index = [0]
            matched = False
        else:
            index, matched = self.find_insertion_point(peak)
        if matched:
            pc = self.peak_clusters[self.find_minimizing_index(peak, index)]
            pc.add(peak)
        else:
            pc = PeakCluster()
            pc.add(peak)
            self.insert_peak_cluster(pc, index)
        self.count += 1

    def insert_peak_cluster(self, peak_cluster, index):
        if index[0] != 0:
            self.peak_clusters.insert(index[0] + 1, peak_cluster)
        else:
            if len(self) == 0:
                new_index = index[0]
            else:
                x = self.peak_clusters[index[0]]
                if x.mz < peak_cluster.mz:
                    new_index = index[0] + 1
                else:
                    new_index = index[0]
            self.peak_clusters.insert(new_index, peak_cluster)

    def cluster_peaks(self, peaks):
        for peak in sorted(peaks, key=lambda x: x.intensity):
            self.handle_peak(peak)

    def compute_probability(self, n_spectra):
        n_spectra = float(n_spectra)
        for pc in self:
            pc.probability = len(pc) / n_spectra


def cluster_peaks(peak_sets, error_tolerance=1e-5, n_spectra=None):
    if n_spectra is None:
        n_spectra = len(peak_sets)
    clusterer = PeakClusterBuilder(error_tolerance)
    clusterer.cluster_peaks(unspool_peaks(peak_sets))
    clusterer.compute_probability(n_spectra)
    return clusterer


def binary_search_with_flag(array, mz, error_tolerance=1e-5):
        lo = 0
        n = hi = len(array)
        while hi != lo:
            mid = (hi + lo) / 2
            x = array[mid]
            err = (x.mz - mz) / mz
            if abs(err) <= error_tolerance:
                i = mid - 1
                # Begin Sweep forward
                while i > 0:
                    x = array[i]
                    err = (x.mz - mz) / mz
                    if abs(err) <= error_tolerance:
                        i -= 1
                        continue
                    else:
                        break
                low_end = i
                i = mid + 1

                # Begin Sweep backward
                while i < n:
                    x = array[i]
                    err = (x.mz - mz) / mz
                    if abs(err) <= error_tolerance:
                        i += 1
                        continue
                    else:
                        break
                high_end = i
                return list(range(low_end, high_end)), True
            elif (hi - lo) == 1:
                return [mid], False
            elif err > 0:
                hi = mid
            elif err < 0:
                lo = mid
        return 0, False


def get_mz_window(peak_set, lo, hi, error_tolerance=1e-5):
    lo_ix, matched = binary_search_with_flag(peak_set, lo, error_tolerance)
    hi_ix, matched = binary_search_with_flag(peak_set, hi, error_tolerance)

    return peak_set[lo_ix[0]:hi_ix[0] + 1]


def top_n_peaks(peak_set, n=5):
    peak_set = sorted(peak_set, key=lambda x: x.intensity, reverse=True)
    try:
        for i in range(n):
            yield peak_set[i]
    except IndexError:
        pass


def filter_peak_set(peak_set, window_width=100, top_n=5):
    peaks = []
    upper_limit = max(peak_set, key=lambda x: x.mz).mz
    current_mz = 0.1
    while current_mz < upper_limit:
        peaks.extend(
            top_n_peaks(get_mz_window(
                peak_set, current_mz, current_mz + window_width),
                top_n))
        current_mz += window_width
    return sorted(peaks, key=lambda x: x.mz)


def average_fitted_peak(peak_cluster, divisor=False):
    if divisor:
        scale = divisor
    else:
        scale = 1
    weighted_mz = peak_cluster.weighted_mz()
    total_intensity = peak_cluster.intensity
    signal_to_noise = sum(p.signal_to_noise * p.intensity for p in peak_cluster) / total_intensity
    fwhm = sum(p.full_width_at_half_max * p.intensity for p in peak_cluster) / total_intensity
    area = sum(p.area * p.intensity for p in peak_cluster) / total_intensity
    return FittedPeak(weighted_mz, total_intensity / scale, signal_to_noise, -1, -1, fwhm, area)


def average_deconvoluted_peak(peak_cluster, divisor=False):
    if divisor:
        scale = divisor
    else:
        scale = 1
    weighted_mz = peak_cluster.weighted_mz()
    total_intensity = peak_cluster.intensity
    neutral_mass = sum(p.neutral_mass * p.intensity for p in peak_cluster) / total_intensity
    most_abundant_mass = sum(p.most_abundant_mass * p.intensity for p in peak_cluster) / total_intensity
    a_to_a2_ratio = sum(p.a_to_a2_ratio * p.intensity for p in peak_cluster) / total_intensity
    average_mass = sum(p.average_mass * p.intensity for p in peak_cluster) / total_intensity
    signal_to_noise = sum(p.signal_to_noise * p.intensity for p in peak_cluster) / total_intensity
    fwhm = sum(p.full_width_at_half_max * p.intensity for p in peak_cluster) / total_intensity
    area = sum(p.area * p.intensity for p in peak_cluster) / total_intensity
    score = sum(p.score * p.intensity for p in peak_cluster) / total_intensity
    charge_counts = Counter(p.charge for p in peak_cluster).items()
    charge = max(charge_counts, key=lambda x: x[1])[0]
    envelope = average_envelope([p.envelope for p in peak_cluster])
    return DeconvolutedPeak(
        neutral_mass=neutral_mass, intensity=total_intensity / scale, charge=charge, signal_to_noise=signal_to_noise,
        full_width_at_half_max=fwhm, index=-1, a_to_a2_ratio=a_to_a2_ratio, most_abundant_mass=most_abundant_mass,
        average_mass=average_mass, score=score, envelope=envelope, mz=weighted_mz, fit=None, chosen_for_msms=False,
        area=area)


def average_envelope(envelopes, error_tolerance=2e-5):
    for position in cluster_peaks(envelopes, error_tolerance):
        yield (position.mz, position.intensity)


def make_peak_index(fitted_peaks):
    ps = PeakSet(fitted_peaks)
    ps._index()
    return ps


def make_deconvoluted_peak_set(deconvoluted_peaks):
    ps = DeconvolutedPeakSet(deconvoluted_peaks)
    ps.reindex()
    return ps



# -*- coding: utf-8 -*-
'''This module defines a collection of isotopic envelope search strategies for
deisotoping and charge state deconvolution. Each strategy is implemented through
a subtype of :class:`~.DeconvoluterBase` and provide a common set of methods for
deconvolving a peak list.
'''
# from .deconvolution import *

from .utils import (
    prepare_peaklist,
    charge_range_,
    ChargeIterator,
    quick_charge,
    drop_placeholders_parallel,
    drop_placeholders,
    first_peak,
    from_fitted_peak,
    count_placeholders,
    mean)

from .base import (
    DeconvoluterBase)

from .exhaustive import (
    ExhaustivePeakSearchDeconvoluterBase,
    PeakDependenceGraphDeconvoluterBase)

from .averagine_based import (
    AveragineDeconvoluterBase,
    AveragineDeconvoluter,
    AveraginePeakDependenceGraphDeconvoluter,
    MultiAveragineDeconvoluterBase,
    MultiAveragineDeconvoluter,
    MultiAveraginePeakDependenceGraphDeconvoluter)

from .composition_list import (
    CompositionListDeconvoluterBase,
    CompositionListDeconvoluter,
    CompositionListPeakDependenceGraphDeconvoluter)

from .api import deconvolute_peaks



import multiprocessing

from .scan_index import ExtendedScanIndex
from .scan_interval_tree import (
    ScanIntervalTree, extract_intervals, make_rt_tree)


n_cores = multiprocessing.cpu_count()


def indexing_iterator(reader, start, end, index):
    try:
        iterator = reader.start_from_scan(index=start, grouped=True)
    except AttributeError:
        if start != 0:
            raise
        iterator = reader
    for scan_bunch in iterator:
        try:
            ix = scan_bunch.precursor.index
        except AttributeError:
            ix = scan_bunch.products[0].index
        if ix > end:
            break
        index.add_scan_bunch(scan_bunch)
        yield scan_bunch


def index_chunk(reader, start, end):
    index = ExtendedScanIndex()
    iterator = indexing_iterator(reader, start, end, index)
    intervals = extract_intervals(iterator)
    return (start, end, index, intervals)


def partition_work(n_items, n_workers, start_index=0):
    chunk_size = int(n_items / n_workers)
    n_items += start_index
    intervals = []
    start = start_index
    intervals.append([start, start + chunk_size])
    start += chunk_size

    while start + chunk_size < (n_items):
        end = start + chunk_size
        intervals.append([start, end])
        start = end

    last = intervals[-1]
    if last[1] > n_items:
        last[1] = n_items
    else:
        last[1] += (n_items - last[1])

    return intervals


class _Indexer(object):
    def __call__(self, payload):
        reader, start, end = payload
        try:
            result = index_chunk(reader, start, end)
        except Exception as e:
            print(reader, start, end, e)
            import traceback
            traceback.print_exc()
            raise e
        return result


class _TaskWrapper(object):
    '''A simple wrapper for a callable to capture the index range for a chunk created by
    :func:`run_task_in_chunks`, so that the start index of the chunk is known.
    '''
    def __init__(self, task):
        self.task = task

    def __call__(self, payload):
        _reader, start, _end = payload
        out = self.task(payload)
        return start, out


def run_task_in_chunks(reader, n_processes=None, n_chunks=None, scan_interval=None, task=None, progress_indicator=None):
    """Run a :class:`~.Callable` `task` over a :class:`~.ScanIterator` in chunks across multiple processes.

    This function breaks apart a :class:`~.ScanIterator`'s scans over `scan_interval`,
    or the whole sequence if not provided.

    Parameters
    ----------
    reader : :class:`~.ScanIterator`
        The set of :class:`~.Scan` objects to operate on
    n_processes : int, optional
        The number of worker processes to use (the default is 4 or the number of cores available,
        whichever is lower)
    n_chunks : int, optional
        The number of chunks to break the scan range into (the default is equal to `n_processes`)
    scan_interval : :class:`tuple` of (:class:`int`, :class:`int`), optional
        The start and stop scan index to apply the task over. If omitted, the entire scan range will
        be used. If either entry is :const:`None`, then the index will be assumed to be the first or
        last scan respectively.
    task : :class:`~.Callable`
        The callable object which will be executed on each chunk in a sub-process. It must take one
        argument, a :class:`tuple` of (`reader`, `start index`, `stop index`), and it must return a pickle-able
        object.
    progress_indicator : :class:`~.Callable`, optional
        A callable object which will be used to report progress as chunks finish processing. It must take
        one argument, a :class:`float` which represents the fraction of all work completed.

    Returns
    -------
    :class:`list`:
        The result of `task` on each chunk of `reader` in index sorted order.

    """
    if n_processes is None:
        n_processes = min(n_cores, 4)
    if task is None or not callable(task):
        raise ValueError("The task must be callable!")
    if scan_interval is None:
        start_scan = 0
        end_scan = len(reader.index)
    else:
        start_scan, end_scan = scan_interval
        if start_scan is None:
            start_scan = 0
        if end_scan is None:
            end_scan = len(reader.index)
    if n_chunks is None:
        n_chunks = n_processes
    n_items = end_scan - start_scan
    pool = multiprocessing.Pool(n_processes)
    scan_ranges = partition_work(n_items, n_chunks, start_scan)
    feeder = ((reader, scan_range[0], scan_range[1]) for scan_range in scan_ranges)
    result = []
    for i, block in enumerate(pool.imap_unordered(_TaskWrapper(task), feeder), 1):
        result.append(block)
        if progress_indicator is not None:
            progress_indicator(i / float(n_chunks))
    pool.close()
    pool.terminate()
    pool.join()
    result.sort(key=lambda x: x[0])
    result = [x[1] for x in result]
    return result


def merge_indices(indices):
    index = indices[0]
    for ind in indices:
        index = index.merge(ind)
    return index


def make_interval_tree(intervals):
    concat = []
    for i in intervals:
        concat.extend(i)
    return ScanIntervalTree(make_rt_tree(concat), None)


def index(reader, n_processes=4, scan_interval=None, progress_indicator=None):
    task = _Indexer()
    # indexing a :class:`ScanIterator` without random access, have to go in sequence
    if not hasattr(reader, 'start_from_scan'):
        reader.make_iterator(grouped=True)
        chunks = [task((reader, 0, len(reader)))]
    else:
        chunks = run_task_in_chunks(
            reader, n_processes, scan_interval=scan_interval, task=task,
            progress_indicator=progress_indicator)
    indices = [chunk[2] for chunk in chunks]
    intervals = [chunk[3] for chunk in chunks]
    index = merge_indices(indices)
    interval_tree = make_interval_tree(intervals)
    return index, interval_tree


def multi_index(readers, n_processes=4, scan_interval=None):
    index_collection = []
    interval_collection = []
    for reader in readers:
        chunks = run_task_in_chunks(
            reader, n_processes, scan_interval=scan_interval, task=_Indexer())
        indices = [chunk[2] for chunk in chunks]
        intervals = [chunk[3] for chunk in chunks]
        index_collection.append(indices)
        interval_collection.append(intervals)



import os

import click

import ms_peak_picker

from ms_peak_picker.scan_filter import parse as parse_filter

import ms_deisotope
from ms_deisotope import MSFileLoader
from ms_deisotope.data_source import RandomAccessScanSource

from ms_deisotope.tools.utils import processes_option, AveragineParamType, is_debug_mode, register_debug_hook
from ms_deisotope.tools.deisotoper import workflow


def check_random_access(loader, start_time, end_time):
    if isinstance(loader, RandomAccessScanSource):
        last_scan = loader[len(loader) - 1]
        last_time = last_scan.scan_time
        start_scan = loader._locate_ms1_scan(
            loader.get_scan_by_time(start_time))
        if end_time > last_time:
            end_time = last_time
        end_scan = loader._locate_ms1_scan(
            loader.get_scan_by_time(end_time))

        start_scan_id = start_scan.id
        end_scan_id = end_scan.id

        start_scan_time = start_scan.scan_time
        end_scan_time = end_scan.scan_time

        loader.reset()
        loader.start_from_scan(start_scan_id, grouped=True)
    else:
        click.secho("The file format provided does not support random"
                    " access, start and end points will be ignored", fg='yellow')
        start_scan_time = 0
        start_scan_id = None

        end_scan_time = float('inf')
        end_scan_id = None
        loader.make_iterator(grouped=True)

    return start_scan_id, start_scan_time, end_scan_id, end_scan_time


def check_if_profile(loader):
    first_bunch = next(loader)
    if first_bunch.precursor is not None:
        is_profile = (first_bunch.precursor.is_profile)
    elif first_bunch.products:
        is_profile = (first_bunch.products[0].is_profile)

    if is_profile:
        click.secho("Spectra are profile")
    else:
        click.secho("Spectra are centroided", fg='yellow')
    return is_profile


@click.command("deisotope", short_help=(
    "Convert raw mass spectra data into deisotoped neutral mass peak lists written to mzML."
    " Can accept mzML, mzXML, MGF with either profile or centroided scans."),
    context_settings=dict(help_option_names=['-h', '--help']))
@click.argument("ms-file", type=click.Path(exists=True))
@click.argument("outfile-path", type=click.Path(writable=True))
@click.option("-a", "--averagine", default=["peptide"],
              type=AveragineParamType(),
              help='Averagine model to use for MS1 scans. Either a name or formula',
              multiple=True)
@click.option("-an", "--msn-averagine", default="peptide",
              type=AveragineParamType(),
              help='Averagine model to use for MS^n scans. Either a name or formula')
@click.option("-s", "--start-time", type=float, default=0.0,
              help='Scan time to begin processing at in minutes')
@click.option("-e", "--end-time", type=float, default=float('inf'),
              help='Scan time to stop processing at in minutes')
@click.option("-c", "--maximum-charge", type=int, default=8,
              help=('Highest absolute charge state to consider'))
@click.option("-n", "--name", default=None,
              help="Name for the sample run to be stored. Defaults to the base name of the input data file")
@click.option("-t", "--score-threshold", type=float, default=workflow.SampleConsumer.MS1_SCORE_THRESHOLD,
              help="Minimum score to accept an isotopic pattern fit in an MS1 scan")
@click.option("-tn", "--msn-score-threshold", type=float, default=workflow.SampleConsumer.MSN_SCORE_THRESHOLD,
              help="Minimum score to accept an isotopic pattern fit in an MS^n scan")
@click.option("-m", "--missed-peaks", type=int, default=3,
              help="Number of missing peaks to permit before an isotopic fit is discarded")
@click.option("-mn", "--msn-missed-peaks", type=int, default=1,
              help="Number of missing peaks to permit before an isotopic fit is discarded in an MSn scan")
@processes_option
@click.option("-b", "--background-reduction", type=float, default=0., help=(
              "Background reduction factor. Larger values more aggresively remove low abundance"
              " signal in MS1 scans."))
@click.option("-bn", "--msn-background-reduction", type=float, default=0., help=(
              "Background reduction factor. Larger values more aggresively remove low abundance"
              " signal in MS^n scans."))
@click.option("-r", '--transform', multiple=True, type=parse_filter, help=(
    "Scan transformations to apply to MS1 scans. May specify more than once."))
@click.option("-rn", '--msn-transform', multiple=True, type=parse_filter, help=(
    "Scan transformations to apply to MS^n scans. May specify more than once."))
@click.option("-v", "--extract-only-tandem-envelopes", is_flag=True, default=False,
              help='Only work on regions that will be chosen for MS/MS')
@click.option("--verbose", is_flag=True, help="Log additional diagnostic information for each scan.")
@click.option("-g", "--ms1-averaging", default=0, type=int, help=(
    "The number of MS1 scans before and after the current MS1 "
    "scan to average when picking peaks."))
@click.option("--ignore-msn", is_flag=True, default=False, help="Ignore MS^n scans")
@click.option("-i", "--isotopic-strictness", default=2.0, type=float)
@click.option("-in", "--msn-isotopic-strictness", default=0.0, type=float)
@click.option("-snr", "--signal-to-noise-threshold", default=1.0, type=float, help=(
    "Signal-to-noise ratio threshold to apply when filtering peaks"))
@click.option("-mo", "--mass-offset", default=0.0, type=float, help=("Shift peak masses by the given amount"))
def deisotope(ms_file, outfile_path, averagine=None, start_time=None, end_time=None, maximum_charge=None,
              name=None, msn_averagine=None, score_threshold=35., msn_score_threshold=10., missed_peaks=1,
              msn_missed_peaks=1, background_reduction=0., msn_background_reduction=0.,
              transform=None, msn_transform=None, processes=4, extract_only_tandem_envelopes=False,
              ignore_msn=False, isotopic_strictness=2.0, ms1_averaging=0,
              msn_isotopic_strictness=0.0, signal_to_noise_threshold=1.0, mass_offset=0.0,
              deconvolute=True, verbose=False):
    '''Convert raw mass spectra data into deisotoped neutral mass peak lists written to mzML.
    '''
    if transform is None:
        transform = []
    if msn_transform is None:
        msn_transform = []

    if (ignore_msn and extract_only_tandem_envelopes):
        click.secho(
            "Cannot use both --ignore-msn and --extract-only-tandem-envelopes",
            fg='red')
        raise click.Abort("Cannot use both --ignore-msn and --extract-only-tandem-envelopes")

    cache_handler_type = workflow.ThreadedMzMLScanStorageHandler
    click.echo("Preprocessing %s" % ms_file)
    minimum_charge = 1 if maximum_charge > 0 else -1
    charge_range = (minimum_charge, maximum_charge)

    loader = MSFileLoader(ms_file)
    (start_scan_id, start_scan_time,
     end_scan_id, end_scan_time) = check_random_access(loader, start_time, end_time)

    is_profile = check_if_profile(loader)

    if name is None:
        name = os.path.splitext(os.path.basename(ms_file))[0]

    if os.path.exists(outfile_path) and not os.access(outfile_path, os.W_OK):
        click.secho("Can't write to output file path", fg='red')
        raise click.Abort()

    click.secho("Initializing %s" % name, fg='green')
    click.echo("from %s (%0.2f) to %s (%0.2f)" % (
        start_scan_id, start_scan_time, end_scan_id, end_scan_time))
    if deconvolute:
        click.echo("charge range: %s" % (charge_range,))

    if is_profile:
        ms1_peak_picking_args = {
            "transforms": [
            ] + list(transform),
            "signal_to_noise_threshold": signal_to_noise_threshold
        }
        if background_reduction:
            ms1_peak_picking_args['transforms'].append(
                ms_peak_picker.scan_filter.FTICRBaselineRemoval(
                    scale=background_reduction, window_length=2))
            ms1_peak_picking_args['transforms'].append(ms_peak_picker.scan_filter.SavitskyGolayFilter())
    else:
        ms1_peak_picking_args = {
            "transforms": [
                ms_peak_picker.scan_filter.FTICRBaselineRemoval(
                    scale=background_reduction, window_length=2),
            ] + list(transform)
        }

    if msn_background_reduction > 0.0:
        msn_peak_picking_args = {
            "transforms": [
                ms_peak_picker.scan_filter.FTICRBaselineRemoval(
                    scale=msn_background_reduction, window_length=2),
            ] + list(msn_transform)
        }
    else:
        msn_peak_picking_args = {
            "transforms": [
            ] + list(msn_transform)
        }

    if mass_offset != 0.0:
        ms1_peak_picking_args['transforms'].append(
            ms_peak_picker.scan_filter.RecalibrateMass(offset=mass_offset))
        msn_peak_picking_args['transforms'].append(
            ms_peak_picker.scan_filter.RecalibrateMass(offset=mass_offset))

    if deconvolute:
        if len(averagine) == 1:
            averagine = averagine[0]
            ms1_deconvoluter_type = ms_deisotope.deconvolution.AveraginePeakDependenceGraphDeconvoluter
        else:
            ms1_deconvoluter_type = ms_deisotope.deconvolution.MultiAveraginePeakDependenceGraphDeconvoluter

        ms1_deconvolution_args = {
            "scorer": ms_deisotope.scoring.PenalizedMSDeconVFitter(score_threshold, isotopic_strictness),
            "max_missed_peaks": missed_peaks,
            "averagine": averagine,
            "truncate_after": workflow.SampleConsumer.MS1_ISOTOPIC_PATTERN_WIDTH,
            "ignore_below": workflow.SampleConsumer.MS1_IGNORE_BELOW,
            "deconvoluter_type": ms1_deconvoluter_type,
            "use_quick_charge": True
        }

        if msn_isotopic_strictness >= 1:
            msn_isotopic_scorer = ms_deisotope.scoring.PenalizedMSDeconVFitter(
                msn_score_threshold, msn_isotopic_strictness)
        else:
            msn_isotopic_scorer = ms_deisotope.scoring.MSDeconVFitter(msn_score_threshold)

        msn_deconvolution_args = {
            "scorer": msn_isotopic_scorer,
            "averagine": msn_averagine,
            "max_missed_peaks": msn_missed_peaks,
            "truncate_after": workflow.SampleConsumer.MSN_ISOTOPIC_PATTERN_WIDTH,
            "ignore_below": workflow.SampleConsumer.MSN_IGNORE_BELOW,
            "use_quick_charge": True
        }
    else:
        ms1_deconvolution_args = None
        msn_deconvolution_args = None

    consumer = workflow.SampleConsumer(
        ms_file,
        ms1_peak_picking_args=ms1_peak_picking_args,
        ms1_deconvolution_args=ms1_deconvolution_args,
        msn_peak_picking_args=msn_peak_picking_args,
        msn_deconvolution_args=msn_deconvolution_args,
        storage_path=outfile_path, sample_name=name,
        start_scan_id=start_scan_id, storage_type=cache_handler_type,
        end_scan_id=end_scan_id, n_processes=processes,
        extract_only_tandem_envelopes=extract_only_tandem_envelopes,
        ignore_tandem_scans=ignore_msn,
        ms1_averaging=ms1_averaging,
        deconvolute=deconvolute,
        verbose=verbose)
    consumer.start()


if is_debug_mode():
    register_debug_hook()

if __name__ == '__main__':
    deisotope.main()



from .task import TaskBase
from .log_utils import LogUtilsMixin, logger, CallInterval, printer as show_message


__all__ = [
    "TaskBase", "LogUtilsMixin", "logger",
    "CallInterval", "show_message"
]



import unittest

from ms_deisotope.peak_dependency_network import intervals
from ms_deisotope.peak_dependency_network.intervals import Interval, IntervalTreeNode


iv0 = Interval(0, 10)
iv1 = Interval(-10, -5)
iv2 = Interval(-10, 0)
iv3 = Interval(-10, 5)
iv4 = Interval(-10, 10)
iv5 = Interval(-10, 20)
iv6 = Interval(0, 20)
iv7 = Interval(5, 20)
iv8 = Interval(10, 20)
iv9 = Interval(15, 20)
iv10 = Interval(-5, 0)


class TestIntervals(unittest.TestCase):
    def test_interval_overlaps_interval(self):
        assert iv0.overlaps(iv0)
        assert not iv0.overlaps(iv1)
        assert iv0.overlaps(iv2)
        assert iv0.overlaps(iv3)
        assert iv0.overlaps(iv4)
        assert iv0.overlaps(iv5)
        assert iv0.overlaps(iv6)
        assert iv0.overlaps(iv7)
        assert iv0.overlaps(iv8)
        assert not iv0.overlaps(iv9)

    def test_contains_interval(self):
        assert iv0.contains_interval(iv0)
        assert not iv0.contains_interval(iv1)
        assert not iv0.contains_interval(iv2)
        assert not iv0.contains_interval(iv3)
        assert not iv0.contains_interval(iv4)
        assert not iv0.contains_interval(iv5)
        assert not iv0.contains_interval(iv6)
        assert not iv0.contains_interval(iv7)
        assert not iv0.contains_interval(iv8)
        assert not iv0.contains_interval(iv9)
        assert not iv0.contains_interval(iv10)

        assert not iv2.contains_interval(iv0)
        assert iv2.contains_interval(iv1)
        assert iv2.contains_interval(iv2)
        assert not iv2.contains_interval(iv3)
        assert not iv2.contains_interval(iv4)
        assert not iv2.contains_interval(iv5)
        assert not iv2.contains_interval(iv6)
        assert not iv2.contains_interval(iv7)
        assert not iv2.contains_interval(iv8)
        assert not iv2.contains_interval(iv9)
        assert iv2.contains_interval(iv10)


class TestIntervalTreeNode(unittest.TestCase):
    def make_intervals(self):
        intervals = [
            Interval(10, 15),
            Interval(7, 12),
            Interval(22, 36),
            Interval(3, 5),
            Interval(4, 9)
        ]
        return intervals

    def test_build(self):
        # generate many redundant intervals to force
        # the interval tree to branch out.
        intervals = self.make_intervals() * 30
        ivt = IntervalTreeNode.build(intervals)

        assert ivt.start == 3
        assert ivt.end == 36

        assert ivt.left is not None
        assert ivt.left.start == 3
        assert ivt.left.end == 12

        assert ivt.right is not None
        assert ivt.right.end == 36
        assert ivt.right.start == 22


if __name__ == '__main__':
    unittest.main()



'''mzML is a standard rich XML-format for raw mass spectrometry data storage.
This module provides :class:`MzMLLoader`, a :class:`~.RandomAccessScanSource`
implementation.

The parser is based on :mod:`pyteomics.mzml`.
'''
import warnings

from six import string_types as basestring

import numpy as np
from pyteomics import mzml
from .common import (
    PrecursorInformation, ScanDataSource,
    ChargeNotProvided, ActivationInformation,
    ScanAcquisitionInformation, ScanEventInformation,
    ScanWindow, IsolationWindow,
    FileInformation, SourceFile, MultipleActivationInformation,
    ScanFileMetadataBase)
from .metadata.activation import (
    supplemental_energy, UnknownDissociation)
from .metadata.instrument_components import (
    InstrumentInformation, ComponentGroup, component,
    instrument_models)
from .metadata.software import Software
from .metadata import file_information
from .metadata import data_transformation
from .metadata.sample import Sample
from .xml_reader import (
    XMLReaderBase, iterparse_until,
    get_tag_attributes, _find_section, in_minutes)


class _MzMLParser(mzml.MzML):
    # we do not care about chromatograms
    _indexed_tags = {'spectrum', }

    def _handle_param(self, element, **kwargs):
        try:
            element.attrib["value"]
        except KeyError:
            element.attrib["value"] = ""
        return super(_MzMLParser, self)._handle_param(element, **kwargs)

    def _determine_array_dtype(self, info):
        dtype = None
        types = {'32-bit float': np.float32, '64-bit float': np.float64,
                 '32-bit integer': np.int32, '64-bit integer': np.int64,
                 'null-terminated ASCII string': np.uint8}
        for t, code in types.items():
            if t in info:
                dtype = code
                del info[t]
                break
        # sometimes it's under 'name'
        else:
            if 'name' in info:
                for t, code in types.items():
                    if t in info['name']:
                        dtype = code
                        info['name'].remove(t)
                        break
        return dtype


def _find_arrays(data_dict, decode=False):
    arrays = dict()
    for key, value in data_dict.items():
        if " array" in key:
            arrays[key] = value.decode() if decode else value
    return arrays


class MzMLDataInterface(ScanDataSource):
    """Provides implementations of all of the methods needed to implement the
    :class:`ScanDataSource` for mzML files. Not intended for direct instantiation.
    """

    def _stray_cvs(self, scan):
        return scan.get("name", [])

    def _scan_arrays(self, scan):
        """Returns raw data arrays for m/z and intensity

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        mz: np.array
            An array of m/z values for this scan
        intensity: np.array
            An array of intensity values for this scan
        """
        try:
            decode = not self._decode_binary
        except AttributeError:
            decode = False
        arrays = _find_arrays(scan, decode=decode)
        try:
            return arrays.pop('m/z array'), arrays.pop("intensity array"), arrays
        except KeyError:
            return np.array([]), np.array([])

    def _get_selected_ion(self, scan):
        try:
            pinfo_dict = scan["precursorList"]['precursor'][0]["selectedIonList"]['selectedIon'][0]
        except KeyError:
            if "precursorList" in scan:
                if 'precursor' in scan['precursorList']:
                    precursor = scan['precursorList']['precursor'][0]
                    if 'selectedIonList' not in precursor:
                        warnings.warn("No selected ions were found for precursor")
                        return None
            return None
        return pinfo_dict

    def _precursor_information(self, scan):
        """Returns information about the precursor ion,
        if any, that this scan was derived form.

        Returns `None` if this scan has no precursor ion

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        PrecursorInformation
        """
        pinfo_dict = self._get_selected_ion(scan)
        if pinfo_dict is None:
            return None
        try:
            precursor_scan_id = scan["precursorList"]['precursor'][0]['spectrumRef']
        except KeyError:
            precursor_scan_id = None
            # only attempt to scan if there are supposed to be MS1 scans in the file
            if self._has_ms1_scans() and self._use_index:
                last_index = self._scan_index(scan) - 1
                current_level = self._ms_level(scan)
                i = 0
                while last_index > 0 and i < 100:
                    prev_scan = self.get_scan_by_index(last_index)
                    if prev_scan.ms_level >= current_level:
                        last_index -= 1
                    else:
                        precursor_scan_id = self._scan_id(prev_scan._data)
                        break
                    i += 1
        pinfo = PrecursorInformation(
            mz=pinfo_dict['selected ion m/z'],
            intensity=pinfo_dict.get('peak intensity', 0.0),
            charge=pinfo_dict.get('charge state', ChargeNotProvided),
            precursor_scan_id=precursor_scan_id,
            source=self,
            product_scan_id=self._scan_id(scan))
        return pinfo

    def _scan_title(self, scan):
        """Returns a verbose name for this scan, if one
        were stored in the file. Usually includes both the
        scan's id string, as well as information about the
        original file and format.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        str
        """
        try:
            return scan["spectrum title"]
        except KeyError:
            return scan["id"]

    def _scan_id(self, scan):
        """Returns the scan's id string, a unique
        identifier for this scan in the context of
        the data file it is recordered in

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        str
        """
        return scan["id"]

    def _scan_index(self, scan):
        """Returns the base 0 offset from the start
        of the data file in number of scans to reach
        this scan.

        If the original format does not natively include
        an index value, this value may be computed from
        the byte offset index.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        int
        """
        return scan['index']

    def _ms_level(self, scan):
        """Returns the degree of exponential fragmentation
        used to produce this scan.

        1 refers to a survey scan of unfragmented ions, 2
        refers to a tandem scan derived from an ms level 1
        ion, and so on.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        int
        """
        return int(scan['ms level'])

    def _scan_time(self, scan):
        """Returns the time in minutes from the start of data
        acquisition to when this scan was acquired.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        float
        """
        try:
            time = scan['scanList']['scan'][0]['scan start time']
            time = in_minutes(time)
            return time
        except KeyError:
            return 0.0

    def _is_profile(self, scan):
        """Returns whether the scan contains profile data (`True`)
        or centroided data (`False`).

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        bool
        """
        return "profile spectrum" in scan or "profile spectrum" in self._stray_cvs(scan)

    def _polarity(self, scan):
        """Returns whether this scan was acquired in positive mode (+1)
        or negative mode (-1).

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        int
        """
        if "positive scan" in scan or "positive scan" in self._stray_cvs(scan):
            return 1
        elif "negative scan" in scan or "negative scan" in self._stray_cvs(scan):
            return -1

    def _activation(self, scan):
        """Returns information about the activation method used to
        produce this scan, if any.

        Returns `None` for MS1 scans

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        ActivationInformation
        """
        try:
            struct = dict(scan['precursorList']['precursor'][0]['activation'])
            activation = None
            activation_methods = []
            for key in tuple(struct):
                if key in ActivationInformation.dissociation_methods:
                    activation_methods.append(key)
                    struct.pop(key)

            if activation_methods:
                activation = activation_methods[0]
                activation_methods = activation_methods[1:]
            else:
                activation = UnknownDissociation
            energy = struct.pop("collision energy", -1)
            if energy == -1:
                energy = struct.pop("activation energy", -1)

            supplemental_energy_ = struct.pop(supplemental_energy, -1)

            if supplemental_energy_ != -1:
                energies = [energy, supplemental_energy_]
            else:
                energies = [energy]

            if len(activation_methods) == 0:
                return ActivationInformation(activation, energy, struct)
            else:
                return MultipleActivationInformation(
                    [activation] + activation_methods, energies, struct)
        except KeyError:
            return None

    def _isolation_window(self, scan):
        try:
            struct = dict(scan['precursorList']['precursor'][0]['isolationWindow'])
        except KeyError:
            return None
        lower = struct.get("isolation window lower offset")
        target = struct.get('isolation window target m/z')
        upper = struct.get("isolation window upper offset")
        if lower is upper is target is None:
            upper = struct.get("isolation window upper limit")
            lower = struct.get("isolation window lower limit")
            if upper is lower is None:
                if target is not None:
                    return IsolationWindow.make_empty(target)
                else:
                    return None
            else:
                target = (upper - lower) / 2 + lower
                upper = upper - target
                lower = target - lower
        elif target is None:
            target = self._precursor_information(scan).mz
        if lower is None:
            lower = 0.0
        if upper is None:
            upper = 0.0
        return IsolationWindow(lower, target, upper)

    def _acquisition_information(self, scan):
        scan_info = {}
        scan_list_struct = scan['scanList']
        combination = "unknown"
        if "no combination" in scan_list_struct:
            combination = "no combination"
        elif "sum of spectra" in scan_list_struct:
            combination = "sum of spectra"
        elif "median of spectra" in scan_list_struct:
            combination = "median of spectra"
        elif "mean of spectra" in scan_list_struct:
            combination = "mean of spectra"
        scan_info['combination'] = combination
        scan_info_scan_list = []
        for scan in scan_list_struct.get("scan", []):
            scan = scan.copy()
            struct = {}
            struct['start_time'] = scan.pop('scan start time', 0)
            struct['drift_time'] = scan.pop('ion mobility drift time', 0)
            struct['injection_time'] = scan.pop("ion injection time", 0)
            windows = []
            for window in scan.pop("scanWindowList", {}).get("scanWindow", []):
                windows.append(ScanWindow(
                    window['scan window lower limit'],
                    window['scan window upper limit']))
            struct['window_list'] = windows
            scan.pop("instrumentConfigurationRef", None)
            struct['traits'] = scan
            scan_info_scan_list.append(ScanEventInformation(**struct))
        scan_info['scan_list'] = scan_info_scan_list
        return ScanAcquisitionInformation(**scan_info)

    def _instrument_configuration(self, scan):
        try:
            scan_list_struct = scan['scanList']
            reference = None
            for scan in scan_list_struct.get("scan", []):
                reference = scan.get("instrumentConfigurationRef")
                if reference is None:
                    continue
            if reference is None:
                reference = self._run_information['defaultInstrumentConfigurationRef']
            config = self._instrument_config[reference]
            return config
        except KeyError:
            return None

    def _annotations(self, scan):
        annot = dict()
        annot['base peak intensity'] = scan.get('base peak intensity')
        annot['base peak m/z'] = scan.get('base peak m/z')
        annot['lowest observed m/z'] = scan.get('lowest observed m/z')
        annot['highest observed m/z'] = scan.get('highest observed m/z')
        try:
            annot['filter string'] = scan['scanList']['scan'][0].get('filter string')
        except KeyError:
            annot['filter string'] = None
        annot['total ion current'] = scan.get('total ion current')
        return annot


class _MzMLMetadataLoader(ScanFileMetadataBase):

    def _collect_reference_groups(self):
        params = next(iterparse_until(self._source, "referenceableParamGroupList", "run"))
        self.reset()
        if params is None:
            return {}
        params = self._source._get_info_smart(params)
        param_groups = params.get("referenceableParamGroup", [])
        by_id = {
            g.pop('id'): g for g in param_groups
        }
        return by_id

    def file_description(self):
        """Read the file metadata and provenance from the ``<fileDescription>`` tag
        if it is present.

        Returns
        -------
        FileInformation
            The description of the file's contents and its sources
        """
        desc = _find_section(self._source, "fileDescription")
        contents = desc.get("fileContent", {})
        if not isinstance(contents, dict):
            contents = {k: '' for k in contents}
        fi = FileInformation(contents, [])
        for sf_data in desc.get('sourceFileList', {}).get("sourceFile", []):
            sf_data = sf_data.copy()
            sf_name = sf_data.pop('name', '')
            sf_location = sf_data.pop('location', '')
            sf_id = sf_data.pop('id', '')
            # incorrectly merged parameters may contaminate the "name" attribute
            if isinstance(sf_name, list):
                temp = sf_name
                sf_name = temp[0]
                for d in temp[1:]:
                    sf_data[d] = ''
            sf = SourceFile(
                sf_name, sf_location, sf_id,
                parameters=sf_data)
            fi.add_file(sf)
        return fi

    def _convert_instrument(self, configuration):
        group_collection = []
        if "componentList" in configuration:
            for category, groups in configuration['componentList'].items():
                if category == 'count':
                    continue
                for group in groups:
                    group = group.copy()
                    order = group.pop('order')
                    parts = [component(key) for key in group]
                    group_collection.append(ComponentGroup(category, parts, order))
        conf_id = configuration['id']
        serial_number = configuration.get("instrument serial number")
        potential_models = [k for k in configuration if instrument_models.get(k) is not None]
        if potential_models:
            instrument_model = potential_models[0]
        else:
            instrument_model = None
        config = InstrumentInformation(conf_id, group_collection, serial_number=serial_number, model=instrument_model)
        return config

    def instrument_configuration(self):
        """Read the instrument configurations settings from the
        ``<instrumentConfigurationList>``

        Returns
        -------
        list of InstrumentConfiguration
            A list of different instrument states that scans may be acquired under
        """
        instrument_info_list = _find_section(self._source, "instrumentConfigurationList").get(
            "instrumentConfiguration", [])
        out = []
        for instrument_info in instrument_info_list:
            if "referenceableParamGroupRef" in instrument_info:
                reference_params = self._collect_reference_groups()
                for group in instrument_info.pop('referenceableParamGroupRef', []):
                    param_group = reference_params[group['ref']]
                    instrument_info.update(param_group)
            out.append(self._convert_instrument(instrument_info))
        return out

    def software_list(self):
        softwares = _find_section(self._source, "softwareList")
        software_list = []
        for software in softwares.get('software', []):
            software_list.append(
                Software(**software))
        return software_list

    def data_processing(self):
        data_processing_list = _find_section(self._source, "dataProcessingList")
        processing_list = []
        for processing_group in data_processing_list.get("dataProcessing", []):
            dpinfo = data_transformation.DataProcessingInformation()
            dpinfo.id = processing_group['id']
            for method_group in processing_group['processingMethod']:
                order = int(method_group.pop("order", 0))
                software_id = method_group.pop("softwareRef", '')
                method = data_transformation.ProcessingMethod(order=order, software_id=software_id)
                for key, value in method_group.items():
                    if key == 'name':
                        key = value
                        value = ''
                    method.add(key, value)
                dpinfo.methods.append(method)
            processing_list.append(dpinfo)
        return processing_list

    def samples(self):
        """Describe the sample(s) used to generate the mass spectrometry
        data contained in this file.

        Returns
        -------
        :class:`list` of :class:`~.Sample`
        """
        sample_list = _find_section(self._source, "sampleList")
        result = []
        for sample_ in sample_list.get("sample", []):
            name = sample_.pop("sampleName", None)
            sample_.setdefault("name", name)
            result.append(Sample(**sample_))
        return result

    def _get_run_attributes(self):
        return get_tag_attributes(self.source, "run")


class MzMLLoader(MzMLDataInterface, XMLReaderBase, _MzMLMetadataLoader):
    """Reads scans from PSI-HUPO mzML XML files. Provides both iterative and
    random access.

    Attributes
    ----------
    source_file: str
        Path to file to read from.
    source: pyteomics.mzml.MzML
        Underlying scan data source
    """

    _parser_cls = _MzMLParser


    def __init__(self, source_file, use_index=True, decode_binary=True, **kwargs):
        self.source_file = source_file
        self._source = self._parser_cls(source_file, read_schema=True, iterative=True,
                                        huge_tree=True, decode_binary=decode_binary,
                                        use_index=use_index)
        self.initialize_scan_cache()
        self._use_index = use_index
        self._decode_binary = decode_binary
        self._run_information = self._get_run_attributes()
        self._instrument_config = {
            k.id: k for k in self.instrument_configuration()
        }
        self._file_description = self.file_description()
        self.make_iterator()

    def _has_msn_scans(self):
        has_msn_key = file_information.MS_MSn_Spectrum in self._file_description
        has_ms1_key = file_information.MS_MS1_Spectrum in self._file_description

        # we know for certain, return True
        if has_msn_key:
            return has_msn_key
        # if we didn't find the key in the file description, if we didn't find
        # the complementary key either, then the metadata may not have been populated
        # so return indeterminate :const:`None`
        if not has_ms1_key:
            return None
        # We know the metadata was populated, so the absence of the key can be trusted
        # to mean this scan type is absent
        return False

    def _has_ms1_scans(self):
        has_msn_key = file_information.MS_MSn_Spectrum in self._file_description
        has_ms1_key = file_information.MS_MS1_Spectrum in self._file_description

        # we know for certain, return True
        if has_ms1_key:
            return has_ms1_key
        # if we didn't find the key in the file description, if we didn't find
        # the complementary key either, then the metadata may not have been populated
        # so return indeterminate :const:`None`
        if not has_msn_key:
            return None
        # We know the metadata was populated, so the absence of the key can be trusted
        # to mean this scan type is absent
        return False

    def has_msn_scans(self):
        return self._has_msn_scans()

    def has_ms1_scans(self):
        return self._has_ms1_scans()

    @property
    def index(self):
        return self._source.index['spectrum']

    def _validate(self, scan):
        return "m/z array" in scan._data

    def _yield_from_index(self, scan_source, start):
        offset_provider = scan_source._offset_index['spectrum']
        keys = list(offset_provider.keys())
        if start is not None:
            if isinstance(start, basestring):
                start = keys.index(start)
            elif isinstance(start, int):
                start = start
            else:
                raise TypeError("Cannot start from object %r" % start)
        else:
            start = 0
        for key in keys[start:]:
            yield scan_source.get_by_id(key)

    def __reduce__(self):
        return self.__class__, (self.source_file, self._use_index, self._decode_binary)
